(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import Morphisms.
Require Import String.

Require Export Ring.
Require Export Setoid.
(* end hide *)

(**
 * FLAM_Algebra: FLAM principal algebra
 *)

(**
 * Syntax
 *)

(**
 ** Names of native principals.
 *)
Inductive name : Type :=
  | Name : string -> name
.

(* Equality of names is decidable. *)
Definition name_dec: forall (n m : name), {n = m} + {n <> m}.
  decide equality.
  apply string_dec.
Defined.

Lemma eq_name: forall (T : Type) n (a b : T),
  (if name_dec n n then a else b) = a.
Proof.
  intros.
  destruct (name_dec n n).
    auto.

    exfalso.
    auto.
Qed.

(**
 ** Syntax definition
 *)
Inductive Prin : Type :=
  | top : Prin
  | bot : Prin
  | native : name -> Prin
  | conj : Prin -> Prin -> Prin
  | disj : Prin -> Prin -> Prin
  | conf : Prin -> Prin
  | integ : Prin -> Prin
  | owns : Prin -> Prin -> Prin
.

Notation "⊤" := (top).
Notation "⊥" := (bot).
Notation "p ∨ q" := (disj p q) (at level 55, right associativity).
Notation "p ∧ q" := (conj p q) (at level 45, right associativity).
Notation "p : q" := (owns p q) (at level 40, no associativity).
Notation "p →" := (conf p) (at level 9).
Notation "p ←" := (integ p) (at level 9).

(* Labels are just principals. *)
Definition Label := Prin.

Notation "p ⊔ q" := ((p ∧ q)→ ∧ (p ∨ q)←)
  (at level 60, right associativity).

Notation "p ⊓ q" := ((p ∨ q)→ ∧ (p ∧ q)←)
  (at level 60, right associativity).

(**
 * Decidability and discrimination
 *)

(* Equality of principals is decidable. *)
Definition Prin_dec: forall (p q : Prin), {p = q} + {p <> q}.
Proof.
  decide equality. apply name_dec.
Qed.

(* Boolean equality on principals. *)
Definition beq_Prin (p q : Prin) : bool :=
  if (Prin_dec p q) then true else false.

Lemma beq_PrinTrue:
  forall p q, beq_Prin p q = true -> p = q.
Proof.
  intros.
  unfold beq_Prin in H.
  destruct (Prin_dec p q) ; auto.
  contradict H. auto.
Qed.

Lemma beq_PrinFalse:
  forall p q, beq_Prin p q = false -> p <> q.
Proof.
  intros.
  unfold beq_Prin in H.
  destruct (Prin_dec p q) ; auto.
  contradict H. auto with bool.
Qed.

Lemma beq_PrinTrueIff:
  forall p q, beq_Prin p q = true <-> p = q.
Proof.
  intros.
  split ; intros.
  { apply beq_PrinTrue. auto. }
  {
    unfold beq_Prin.
    destruct (Prin_dec p q) ; auto.
  }
Qed.

Lemma beq_PrinFalseIff:
  forall p q, beq_Prin p q = false <-> p <> q.
Proof.
  intros.
  split ; intros.
  { apply beq_PrinFalse. auto. }
  {
    unfold beq_Prin.
    destruct (Prin_dec p q) ; tauto.
  }
Qed.


(**
 * Principal equivalence
 *)
Reserved Infix "==" (at level 70, no associativity).

Inductive prinEq : relation Prin :=
    (* Reflexive transitive symmetric closure *)
  | refl:  forall p,                          p == p
  | symm:  forall p q,              q == p -> p == q
  | trans: forall p q r,  p == q -> q == r -> p == r

    (* Lattice commutativity laws *)
  | conjComm: forall p q,                 p ∧ q == q ∧ p
  | disjComm: forall p q,                 p ∨ q == q ∨ p

    (* Lattice associativity laws *)
  | conjAssoc: forall p q r,        (p ∧ q) ∧ r == p ∧ (q ∧ r)
  | disjAssoc: forall p q r,        (p ∨ q) ∨ r == p ∨ (q ∨ r)

    (* Lattice absorption laws *)
  | conjDisjAbsorb: forall p q,     p ∧ (p ∨ q) == p
  | disjConjAbsorb: forall p q,     p ∨ (p ∧ q) == p

    (* Lattice idempotence laws *)
  | conjIdemp: forall p,                  p ∧ p == p
  | disjIdemp: forall p,                  p ∨ p == p

    (* Lattice identity laws *)
  | conjIdent: forall p,                  p ∧ ⊥ == p
  | disjIdent: forall p,                  p ∨ ⊤ == p
  | conjTop:   forall p,                  p ∧ ⊤ == ⊤
  | disjBot:   forall p,                  p ∨ ⊥ == ⊥

    (* Lattice distributivity law *)
  | conjDistDisj:  forall p q r,    p ∧ (q ∨ r) == (p ∧ q) ∨ (p ∧ r)

    (* Authority projections, property 3: distribution over conjunctions *)
  | conjConf:    forall p q,           (p ∧ q)→ == p→ ∧ q→
  | conjInteg:   forall p q,           (p ∧ q)← == p← ∧ q←
  | ownDistConj: forall p q r,        (p ∧ q):r == p:r ∧ q:r

    (* Authority projections, property 4: distribution over disjunctions *)
  | disjConf:    forall p q,           (p ∨ q)→ == p→ ∨ q→
  | disjInteg:   forall p q,           (p ∨ q)← == p← ∨ q←
  | ownDistDisj: forall p q r,        (p ∨ q):r == p:r ∨ q:r

    (* Authority projections, property 5: idempotence *)
  | confIdemp:  forall p,                   p→→ == p→
  | integIdemp: forall p,                   p←← == p←
  | ownIdemp:   forall p q,             (p:q):q == p:q

    (* Basis projections, properties 2-3 *)
  | confInteg:      forall p,                p→← == ⊥
  | integConf:      forall p,                p←→ == ⊥
  | confDisjInteg:  forall p q,          p→ ∨ q← == ⊥

    (* Ownership projections, properties 3-6 *)
  | ownsSelf:    forall p,                  p:p == p
  | ownsBot:     forall p,                  p:⊥ == ⊥
  | conjDistOwn: forall p q r,        p:(q ∧ r) == p:q ∧ p:r
  | disjDistOwn: forall p q r,        p:(q ∨ r) == p:q ∨ p:r

    (* Ownership projections, property 7 *)
  | ownConfAssoc:  forall p q,             p:q→ == (p:q)→
  | ownIntegAssoc: forall p q,             p:q← == (p:q)←

    (* Ownership projections, property 8 *)
  | ownConf:  forall p q,                  p→:q == (p:q)→
  | ownInteg: forall p q,                  p←:q == (p:q)←

    (* Admitted equivalences *)
  | confIntegBasis: forall p,            p→ ∧ p← == p
  | botConf:                                  ⊥→ == ⊥
  | botInteg:                                 ⊥← == ⊥
  | botOwn:         forall p,                ⊥:p == ⊥

    (* Substitutions *)
  | conjSubst:  forall p p' q q', p == p' -> q == q' -> p ∧ q == p' ∧ q'
  | disjSubst:  forall p p' q q', p == p' -> q == q' -> p ∨ q == p' ∨ q'
  | confSubst:  forall p p',      p == p' ->               p→ == p'→
  | integSubst: forall p p',      p == p' ->               p← == p'←
  | ownsSubst:  forall p p' q q', p == p' -> q == q' ->   p:q == p':q'

where "p == q" := (prinEq p q).

Tactic Notation "autoalg" := (autorewrite with prinalg ; auto with prinalg).
Tactic Notation "eautoalg" := (autorewrite with prinalg ; eauto with prinalg).

Hint Resolve refl : prinalg.
Hint Resolve symm : prinalg.

Hint Resolve conjComm : prinalg.
Hint Resolve disjComm : prinalg.

Hint Resolve conjAssoc : prinalg.
Hint Resolve disjAssoc : prinalg.

Hint Resolve disjConjAbsorb : prinalg. Hint Rewrite disjConjAbsorb : prinalg.
Hint Resolve conjDisjAbsorb : prinalg. Hint Rewrite conjDisjAbsorb : prinalg.

Hint Resolve conjIdemp : prinalg.      Hint Rewrite conjIdemp : prinalg.
Hint Resolve disjIdemp : prinalg.      Hint Rewrite disjIdemp : prinalg.

Hint Resolve conjIdent : prinalg.      Hint Rewrite conjIdent : prinalg.
Hint Resolve disjIdent : prinalg.      Hint Rewrite disjIdent : prinalg.
Hint Resolve conjTop : prinalg.        Hint Rewrite conjTop : prinalg.
Hint Resolve disjBot : prinalg.        Hint Rewrite disjBot : prinalg.

Hint Resolve conjDistDisj : prinalg.

Hint Resolve conjConf : prinalg.
Hint Resolve conjInteg : prinalg.
Hint Resolve ownDistConj : prinalg.    Hint Rewrite ownDistConj : prinalg.

Hint Resolve disjConf : prinalg.
Hint Resolve disjInteg : prinalg.
Hint Resolve ownDistDisj : prinalg.    Hint Rewrite ownDistDisj : prinalg.

Hint Resolve confIdemp : prinalg.      Hint Rewrite confIdemp : prinalg.
Hint Resolve integIdemp : prinalg.     Hint Rewrite integIdemp : prinalg.
Hint Resolve ownIdemp : prinalg.       Hint Rewrite ownIdemp : prinalg.

Hint Resolve confInteg : prinalg.      Hint Rewrite confInteg : prinalg.
Hint Resolve integConf : prinalg.      Hint Rewrite integConf : prinalg.
Hint Resolve confDisjInteg : prinalg.  Hint Rewrite confDisjInteg : prinalg.

Hint Resolve ownsSelf : prinalg.       Hint Rewrite ownsSelf : prinalg.
Hint Resolve ownsBot : prinalg.        Hint Rewrite ownsBot : prinalg.
Hint Resolve conjDistOwn : prinalg.    Hint Rewrite conjDistOwn : prinalg.
Hint Resolve disjDistOwn : prinalg.    Hint Rewrite disjDistOwn : prinalg.

Hint Resolve ownConfAssoc : prinalg.
Hint Resolve ownIntegAssoc : prinalg.

Hint Resolve ownConf : prinalg.        Hint Rewrite ownConf : prinalg.
Hint Resolve ownInteg : prinalg.       Hint Rewrite ownInteg : prinalg.

Hint Resolve confIntegBasis : prinalg. Hint Rewrite confIntegBasis : prinalg.
Hint Resolve botConf : prinalg.        Hint Rewrite botConf : prinalg.
Hint Resolve botInteg : prinalg.       Hint Rewrite botInteg : prinalg.
Hint Resolve botOwn : prinalg.         Hint Rewrite botOwn : prinalg.


(* prinEq is an equivalence relation. *)
Lemma prinEq_refl: Reflexive prinEq.
Proof.
  intro. apply refl.
Qed.

Hint Resolve prinEq_refl : prinalg.

Lemma prinEq_symm: Symmetric prinEq.
Proof.
  intro. intros. apply symm. auto.
Qed.

Hint Immediate prinEq_symm : prinalg.

Lemma prinEq_trans: Transitive prinEq.
Proof.
  intro. intros. eapply trans ; eauto.
Qed.

Hint Resolve prinEq_trans : prinalg.

(* prinEq is a setoid equality. *)
Instance PrinSetoid : Equivalence prinEq.
Proof.
  split ; eautoalg.
Qed.

(**
 ** Setoid compatability
 *)
Instance conjProper : Proper (prinEq ==> prinEq ==> prinEq) conj.
Proof.
  intro. intros. intro. intros. apply conjSubst ; auto.
Qed.

Instance disjProper : Proper (prinEq ==> prinEq ==> prinEq) disj.
Proof.
  intro. intros. intro. intros. apply disjSubst ; auto.
Qed.

Instance confProper : Proper (prinEq ==> prinEq) conf.
Proof.
  intro. intros. apply confSubst. auto.
Qed.

Instance integProper : Proper (prinEq ==> prinEq) integ.
Proof.
  intro. intros. apply integSubst. auto.
Qed.

Instance ownsProper : Proper (prinEq ==> prinEq ==> prinEq) owns.
Proof.
  intro. intros. intro. intros. apply ownsSubst ; auto.
Qed.

(**
 * Derived properties
 *)

(**
 ** Commutativity.
 *)
Lemma joinComm: forall p q, p ⊔ q == q ⊔ p.
Proof.
  intros.
  repeat rewrite (conjComm p q), (disjComm p q).
  autoalg.
Qed.

Lemma meetComm: forall p q, p ⊓ q == q ⊓ p.
Proof.
  intros.
  repeat rewrite (conjComm p q), (disjComm p q).
  autoalg.
Qed.

(**
 ** Associativity.
 *)

Lemma conjAssocRev: forall p q r, p ∧ (q ∧ r) == (p ∧ q) ∧ r.
Proof.
  intros.
  rewrite conjAssoc.
  autoalg.
Qed.

Hint Resolve conjAssocRev : prinalg.

Lemma disjAssocRev: forall p q r, p ∨ (q ∨ r) == (p ∨ q) ∨ r.
Proof.
  intros.
  rewrite disjAssoc.
  autoalg.
Qed.

Hint Resolve disjAssocRev : prinalg.

(**
 ** Distributivity.
 *)

Lemma conjRDistDisj: forall p q r, (p ∨ q) ∧ r == (p ∧ r) ∨ (q ∧ r).
Proof.
  intros.
  rewrite conjComm, conjDistDisj.
  repeat rewrite (conjComm r _).
  autoalg.
Qed.

Hint Resolve conjRDistDisj : prinalg.

Lemma disjLDistConj: forall p q r, p ∨ (q ∧ r) == (p ∨ q) ∧ (p ∨ r).
(* begin hide *)
Proof.
  (*
   * adapted from
   * https://proofwiki.org/wiki/Equivalence_of_Distributive_Lattice_Definitions
   *)
  intros.
  rewrite conjDistDisj.
  rewrite (conjComm _ p).
  rewrite conjDisjAbsorb.
  repeat rewrite conjRDistDisj.
  rewrite <- disjAssoc.
  autoalg.
Qed.
(* end hide *)

Hint Resolve disjLDistConj : prinalg.

Lemma disjRDistConj: forall p q r, (p ∧ q) ∨ r == (p ∨ r) ∧ (q ∨ r).
Proof.
  intros.
  rewrite disjComm, disjLDistConj.
  repeat rewrite (disjComm r _).
  autoalg.
Qed.

Hint Resolve disjRDistConj : prinalg.

(**
 ** Identities.
 *)

(**
 *** Authorization lattice.
 *)

Lemma conjIdentL: forall p, ⊥ ∧ p == p.
Proof.
  intros. rewrite conjComm. autoalg.
Qed.

Hint Resolve conjIdentL : prinalg.  Hint Rewrite conjIdentL : prinalg.

Lemma disjIdentL: forall p, ⊤ ∨ p == p.
Proof.
  intros. rewrite disjComm. autoalg.
Qed.

Hint Resolve disjIdentL : prinalg.  Hint Rewrite disjIdentL : prinalg.

Lemma conjTopL: forall p, ⊤ ∧ p == ⊤.
Proof.
  intros. rewrite conjComm. autoalg.
Qed.

Hint Resolve conjTopL : prinalg.    Hint Rewrite conjTopL : prinalg.

Lemma disjBotL: forall p, ⊥ ∨ p == ⊥.
Proof.
  intros. rewrite disjComm. autoalg.
Qed.

Hint Resolve disjBotL : prinalg.    Hint Rewrite disjBotL : prinalg.

Lemma confIntegBasis': forall p, p← ∧ p→ == p.
Proof.
  intros.
  rewrite conjComm.
  autoalg.
Qed.

Hint Resolve confIntegBasis' : prinalg.

Lemma integDisjConf: forall p q, p← ∨ q→ == ⊥.
Proof.
  intros.
  rewrite disjComm.
  autoalg.
Qed.

Hint Resolve integDisjConf : prinalg.   Hint Rewrite integDisjConf : prinalg.

Lemma disjConfL: forall p q, p→ ∨ q == p→ ∨ q→.
Proof.
  intros.
  setoid_replace (q) with (q→ ∧ q←) at 1 by autoalg.
  rewrite disjLDistConj.
  autoalg.
Qed.

Hint Resolve disjConfL : prinalg.

Lemma disjConfR: forall p q, p ∨ q→ == p→ ∨ q→.
Proof.
  intros.
  rewrite disjComm.
  rewrite disjConfL.
  autoalg.
Qed.

Hint Resolve disjConfR : prinalg.

Lemma disjIntegR: forall p q, p ∨ q← == p← ∨ q←.
Proof.
  intros.
  setoid_replace (p) with (p→ ∧ p←) at 1 by autoalg.
  rewrite disjRDistConj.
  autoalg.
Qed.

Hint Resolve disjIntegR : prinalg.

Lemma disjIntegL: forall p q, p← ∨ q == p← ∨ q←.
Proof.
  intros.
  rewrite disjComm.
  rewrite disjIntegR.
  autoalg.
Qed.

Hint Resolve disjIntegL : prinalg.

Lemma conjOwnConfL: forall p, p→ ∧ p == p.
Proof.
  intros.
  setoid_replace p with (p→ ∧ p←) by autoalg.
  rewrite conjAssocRev.
  autoalg.
Qed.

Hint Resolve conjOwnConfL : prinalg.    Hint Rewrite conjOwnConfL : prinalg.

Lemma conjOwnConfR: forall p, p ∧ p→ == p.
Proof.
  intros.
  rewrite conjComm.
  apply conjOwnConfL.
Qed.

Hint Resolve conjOwnConfR : prinalg.    Hint Rewrite conjOwnConfR : prinalg.

Lemma conjOwnIntegL: forall p, p← ∧ p == p.
Proof.
  intros.
  setoid_replace p with (p→ ∧ p←) by autoalg.
  rewrite conjAssocRev.
  autoalg.
  setoid_replace (p← ∧ p→) with (p→ ∧ p←) by apply conjComm.
  rewrite conjAssoc.
  autoalg.
Qed.

Hint Resolve conjOwnIntegL : prinalg.   Hint Rewrite conjOwnIntegL : prinalg.

Lemma conjOwnIntegR: forall p, p ∧ p← == p.
Proof.
  intros.
  rewrite conjComm.
  apply conjOwnIntegL.
Qed.

Hint Resolve conjOwnIntegR : prinalg.   Hint Rewrite conjOwnIntegR : prinalg.

Lemma disjOwnConfL: forall p, p→ ∨ p == p→.
Proof.
  intros.
  setoid_replace p with (p→ ∧ p←) by autoalg.
  rewrite disjLDistConj.
  autoalg.
Qed.

Hint Resolve disjOwnConfL : prinalg.    Hint Rewrite disjOwnConfL : prinalg.

Lemma disjOwnConfR: forall p, p ∨ p→ == p→.
Proof.
  intros.
  rewrite disjComm.
  apply disjOwnConfL.
Qed.

Hint Resolve disjOwnConfR : prinalg.    Hint Rewrite disjOwnConfR : prinalg.

Lemma disjOwnIntegL: forall p, p← ∨ p == p←.
Proof.
  intros.
  setoid_replace p with (p→ ∧ p←) by autoalg.
  rewrite disjLDistConj.
  autoalg.
Qed.

Hint Resolve disjOwnIntegL : prinalg.   Hint Rewrite disjOwnIntegL : prinalg.

Lemma disjOwnIntegR: forall p, p ∨ p← == p←.
Proof.
  intros.
  rewrite disjComm.
  apply disjOwnIntegL.
Qed.

Hint Resolve disjOwnIntegR : prinalg.   Hint Rewrite disjOwnIntegR : prinalg.

Lemma conjPermute: forall p q r, (p ∧ q) ∧ r == (p ∧ r) ∧ q.
Proof.
  intros.
  rewrite conjAssoc.
  setoid_replace (q ∧ r) with (r ∧ q) by apply conjComm.
  autoalg.
Qed.

Hint Resolve conjPermute : prinalg.

Lemma disjPermute: forall p q r, (p ∨ q) ∨ r == (p ∨ r) ∨ q.
Proof.
  intros.
  rewrite disjAssoc.
  setoid_replace (q ∨ r) with (r ∨ q) by apply disjComm.
  autoalg.
Qed.

Hint Resolve disjPermute : prinalg.

Lemma conjPermute2in4: forall p q r s, (p ∧ q) ∧ (r ∧ s) == (p ∧ r) ∧ (q ∧ s).
Proof.
  intros.
  rewrite conjPermute.
  rewrite conjAssoc.
  rewrite conjPermute.
  rewrite conjAssoc.
  autoalg.
Qed.

Hint Resolve conjPermute2in4 : prinalg.

Lemma disjPermute2in4: forall p q r s, (p ∨ q) ∨ (r ∨ s) == (p ∨ r) ∨ (q ∨ s).
Proof.
  intros.
  rewrite disjPermute.
  rewrite disjAssoc.
  rewrite disjPermute.
  rewrite disjAssoc.
  autoalg.
Qed.

Hint Resolve disjPermute2in4 : prinalg.

(**
 *** Information-flow lattice.
 *)
Lemma joinTopConf: forall p, (⊤→ ⊔ p) == ⊤→.
Proof.
  intros.
  rewrite conjConf.
  rewrite disjInteg.
  autoalg.
  rewrite <- conjConf.
  autoalg.
Qed.

Hint Resolve joinTopConf : prinalg.   Hint Rewrite joinTopConf : prinalg.

Lemma joinTopConf': forall p, (p ⊔ ⊤→) == ⊤→.
Proof.
  intros.
  rewrite joinComm.
  apply joinTopConf.
Qed.

Hint Resolve joinTopConf' : prinalg.  Hint Rewrite joinTopConf' : prinalg.

Lemma meetTopInteg: forall p, (⊤← ⊓ p) == ⊤←.
Proof.
  intros.
  rewrite conjInteg.
  rewrite disjConf.
  autoalg.
  rewrite <- conjInteg.
  autoalg.
Qed.

Hint Resolve meetTopInteg : prinalg.  Hint Rewrite meetTopInteg : prinalg.

Lemma meetTopInteg': forall p, (p ⊓ ⊤←) == ⊤←.
Proof.
  intros.
  rewrite meetComm.
  apply meetTopInteg.
Qed.

Hint Resolve meetTopInteg' : prinalg. Hint Rewrite meetTopInteg' : prinalg.

Lemma confJoin: forall p q, (p ⊔ q)→ == (p ∧ q)→.
Proof.
  intros.
  rewrite conjConf.
  eautoalg.
Qed.

Hint Resolve confJoin : prinalg.      Hint Rewrite confJoin : prinalg.

Lemma integJoin: forall p q, (p ⊔ q)← == (p ∨ q)←.
Proof.
  intros.
  rewrite conjInteg.
  eautoalg.
Qed.

Hint Resolve integJoin : prinalg.     Hint Rewrite integJoin : prinalg.

Lemma confMeet: forall p q, (p ⊓ q)→ == (p ∨ q)→.
Proof.
  intros.
  rewrite conjConf.
  eautoalg.
Qed.

Hint Resolve confMeet : prinalg.      Hint Rewrite confMeet : prinalg.

Lemma integMeet: forall p q, (p ⊓ q)← == (p ∧ q)←.
Proof.
  intros.
  rewrite conjInteg.
  eautoalg.
Qed.

Hint Resolve integMeet : prinalg.     Hint Rewrite integMeet : prinalg.


(**
 * Semi-ring definitions.
 *)

(*
 * Define a semi-ring theory with ∧ as addition (with identity ⊥) and ∨ as
 * multiplication (with identity ⊤), so we obtain a join-normal form.
 *)
Definition prin_semiring :=
    mk_srt bot top conj disj prinEq conjIdentL conjComm conjAssocRev disjIdentL
      disjBotL disjComm disjAssocRev disjRDistConj.

Add Ring PrinSemiRing : prin_semiring.
