(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import Equalities.
Require Import FiniteSets.
Require Import FLAM_Algebra.
Require Import FLAM_DelSet.
Require Import List.
Require Import Morphisms.
Require Export Setoid.
(* end hide *)

(*
 * A DelUse is a delegation used in a derivation, paired with a list of node
 * names, representing the path of nodes through which the delegation was
 * forwarded. The head of the list is the node directly contacted by the local
 * node at the root of the derivation, and the end of the list is the
 * delegation's origin (i.e., the node hosting the delegation). Locally used
 * delegations (i.e., those not used in the derivation of an instance of the
 * second premise of the Fwd rule) are paired with the empty list.
 *)
Module DelUse <: MiniDecidableType.
  Inductive T :=
  | make : Del.t -> list name -> T.

  Definition t := T.
  Definition eq_dec : forall (d1 d2 : t), {d1 = d2} + {d1 <> d2}.
    decide equality.
    { decide equality. apply name_dec. }
    { apply Del.eq_dec. }
  Qed.

  Definition init d := make d nil.

  Definition getDel d :=
    match d with
    | make d0 _ => d0
    end.

  Definition getPath d :=
    match d with
    | make _ path => path
    end.

  Definition getHost d :=
    match d with
    | make (DelT.make _ _ _ rn) _ => rn
    end.

  (* Adds a node to the forwarding path. *)
  Definition transfer d rn :=
    match d with
    | make d0 path => make d0 (rn :: path)
    end.

  Definition disjPath_f (n : name) (accum : Prin) :=
    (native n) ∨ accum.

  (* The disjunction of all principals on the forwarding path. *)
  Definition disjPath d :=
    fold_right disjPath_f ⊤ (getPath d).

  Lemma disjPathInit:
    forall d, disjPath (init d) = ⊤.
  Proof.
    intros. unfold disjPath. simpl. auto.
  Qed.

  Lemma disjPathTransfer:
    forall d rn, disjPath (transfer d rn) = (native rn) ∨ disjPath d.
  Proof.
    intros.
    destruct d as [d path]. simpl.
    unfold disjPath. simpl.
    unfold disjPath_f at 1. auto.
  Qed.
End DelUse.

(**
 * Delegation-use sets.
 *)

(*
 * A finite set of DelUses.
 *)
Module DelUseSet.
  Module Raw := FiniteSet DelUse.
  Include Raw.

  Notation "A =D= B" := (sameSet A B) (at level 70, no associativity).

  Definition DelIn (d : Del.t) (S : t) :=
    exists du, In du S /\ d = DelUse.getDel du.

  Instance DelInProper:
    Proper (eq ==> sameSet ==> iff) DelIn.
  Proof.
    do 2 (intro ; intros).
    subst y.
    unfold DelIn.
    split;
      intro;
      destruct H as [? [? ?]];
      [rewrite H0 in * | rewrite <- H0 in *];
      eauto.
  Qed.

  Lemma emptyNotDelIn:
    forall d, ~ DelIn d empty.
  Proof.
    intros. intro.
    unfold DelIn in *.
    destruct H as [? [? ?]].
    contradict H.
    apply emptyIsEmpty.
  Qed.

  (*
   * Creates a DelUseSet with a singleton local DelUse.
   *)
  Definition init (d : Del.t) : t := singleton (DelUse.init d).

  Definition transfer_f
      (rn : name) (d : DelUse.t) (accum : t) :=
    union accum (singleton (DelUse.transfer d rn)).

  Instance transfer_fProper rn:
    Proper (eq ==> sameSet ==> sameSet) (transfer_f rn).
  Proof.
    intro. do 2 (intro ; intros).
    unfold transfer_f.
    rewrite H, H0.
    auto with flam.
  Qed.

  Lemma transfer_fTranspose:
    forall rn, SetoidList.transpose sameSet (transfer_f rn).
  Proof.
    intro. intro. intros.
    unfold transfer_f.
    remember (singleton (DelUse.transfer y rn)) as S1.
    remember (singleton (DelUse.transfer x rn)) as S2.
    rewrite unionAssoc.
    rewrite (unionComm S1 S2).
    rewrite <- unionAssoc.
    auto with flam.
  Qed.

  Lemma transfer_fIdempotent:
    forall rn, idempotent t sameSet (transfer_f rn).
  Proof.
    intros. intro. intros.
    unfold transfer_f.
    apply unionAbsorbs.
    auto with flam.
  Qed.

  (*
   * Transfers a DelUseSet obtained from a given host.
   *)
  Definition transfer (D : t) (rn : name) :=
    fold t (transfer_f rn) D empty.

  Instance transferProper:
    Proper (sameSet ==> eq ==> sameSet) transfer.
  Proof.
    do 2 (intro ; intros).
    subst y0.
    unfold transfer.
    pose proof foldProper _ _ (transfer_f x0) sameSetSetoid
      (transfer_fProper x0) (transfer_fTranspose x0).
    rewrite H.
    split ; intros ; auto.
  Qed.

  Lemma transferEmpty:
    forall rn D, Empty D -> transfer D rn = empty.
  Proof.
    intros. unfold transfer. rewrite foldEmpty ; auto.
  Qed.

  Lemma transferCorrect1:
    forall D d rn,
      In d D <-> In (DelUse.transfer d rn) (transfer D rn).
  Proof.
    induction D using set_ind ; intros ; split ; intro.
    { contradict H0. auto. }
    { contradict H0. rewrite transferEmpty ; auto. apply emptyIsEmpty. }
    {
      unfold transfer.
      rewrite foldAdd ; eauto.
      {
        fold (transfer D1 rn).
        unfold transfer_f.
        apply unionIntro.
        destruct (DelUse.eq_dec x d).
        { right. subst. apply singletonIntro. auto. }
        {
          left.
          apply IHD1.
          destruct (H0 d) as [Hadd _].
          destruct (Hadd H1) ; tauto.
        }
      }
      { apply sameSetSetoid. }
      { apply transfer_fProper. }
      { apply transfer_fTranspose. }
    }
    {
      apply H0.
      unfold transfer in H1.
      rewrite foldAdd with (eqB := sameSet) in H1 ; eauto.
      {
        fold (transfer D1 rn) in H1.
        unfold transfer_f in H1.
        destruct (unionInv _ _ _ H1).
        { right. eapply IHD1. eauto. }
        {
          left.
          pose proof (singletonInv _ _ H2) as Heq.
          destruct x, d.
          simpl in Heq.
          inversion Heq.
          auto.
        }
      }
      { apply sameSetSetoid. }
      { apply transfer_fProper. }
      { apply transfer_fTranspose. }
    }
  Qed.

  Lemma transferCorrect2:
    forall D d rn,
         In d (transfer D rn)
      -> exists d0, In d0 D /\ d = DelUse.transfer d0 rn.
  Proof.
    induction D using set_ind ; intros.
    { contradict H0. rewrite transferEmpty ; auto. apply emptyIsEmpty. }
    {
      unfold transfer in H1.
      rewrite foldAdd with (eqB := sameSet) in H1 ; eauto.
      {
        fold (transfer D1 rn) in H1.
        unfold transfer_f in H1.
        destruct (unionInv _ _ _ H1).
        {
          destruct (IHD1 _ _ H2) as [? [? ?]].
          eexists. split ; eauto.
          apply H0. eauto.
        }
        {
          pose proof singletonInv _ _ H2.
          eexists. split ; eauto.
          apply H0. eauto.
        }
      }
      { apply sameSetSetoid. }
      { apply transfer_fProper. }
      { apply transfer_fTranspose. }
    }
  Qed.

  Lemma initInv:
    forall x y, In x (init y) -> x = (DelUse.init y).
  Proof.
    intros.
    unfold init in H.
    pose proof (singletonInv _ _ H).
    auto.
  Qed.

  Hint Resolve initInv : flam.

  Lemma initIntro:
    forall x, In (DelUse.init x) (init x).
  Proof.
    intros.
    unfold init.
    apply singletonIntro.
    auto.
  Qed.

  Hint Resolve initIntro : flam.

  Lemma initNotEmpty:
    forall x, ~ Empty (init x).
  Proof.
    intros. intro.
    apply (H (DelUse.init x)).
    apply initIntro.
  Qed.

  (**
   ** Properties
   *)

  Lemma delInTransferInv : forall d rn D,
    DelIn d (transfer D rn) -> DelIn d D.
  Proof.
    intros.
    unfold DelIn in *.
    destruct H as [? [? ?]].

    destruct (transferCorrect2 _ _ _ H) as [y [? ?]].
    eexists. split ; eauto.
    subst. destruct y. auto.
  Qed.

  Lemma unionInvDel : forall d D1 D2,
    DelIn d (union D1 D2) -> DelIn d D1 \/ DelIn d D2.
  Proof.
    intros.
    unfold DelIn in *.
    destruct H as [? [? ?]].
    destruct (unionInv _ _ _ H);
      [left|right];
      eauto.
  Qed.

  Lemma unionInvDel3 : forall d D1 D2 D3,
    DelIn d (union D1 (union D2 D3))
      -> DelIn d D1 \/ (DelIn d D2 \/ DelIn d D3).
  Proof.
    intros.
    destruct (unionInvDel _ _ _ H) ; auto.
    right.
    apply unionInvDel. auto.
  Qed.

  Lemma transferIsEmpty:
    forall D rn, transfer D rn =D= empty -> D =D= empty.
  Proof.
    intros. intro. split ; intros.
    {
      assert (D =D= union (singleton x) D) as Hrewrite.
      {
        symmetry. rewrite unionComm. apply unionAbsorbs.
        intro. intros. pose proof (singletonInv _ _ H1). subst. auto.
      }

      rewrite Hrewrite in H.
      unfold transfer in H.
      rewrite foldUnion with (eqB := sameSet) in H.
      {
        fold (transfer D rn) in H.
        rewrite foldSingleton with (eqB := sameSet) in H.
        {
          unfold transfer_f in H.
          remember (DelUse.transfer x rn) as x'.
          destruct (H x') as [? _].
          assert (In x' (union (transfer D rn) (singleton x'))).
          { apply unionIntro. right. apply singletonIntro. auto. }
          pose proof (H1 H2) as Hcontra.
          contradict Hcontra.
          apply emptyIsEmpty.
        }
        { apply sameSetSetoid. }
        { apply transfer_fProper. }
        { apply transfer_fTranspose. }
      }
      { apply sameSetSetoid. }
      { apply transfer_fProper. }
      { apply transfer_fTranspose. }
      { apply transfer_fIdempotent. }
    }
    { contradict H0. apply emptyIsEmpty. }
  Qed.

  Lemma transferMonotonic:
    forall D1 D2 rn,
      included D1 D2 -> included (transfer D1 rn) (transfer D2 rn).
  Proof.
    intros.
    intro. intros.
    destruct (transferCorrect2 _ _ _ H0) as [? [? ?]].
    subst.
    apply transferCorrect1.
    apply H. auto.
  Qed.

  Hint Resolve transferMonotonic : flam.

End DelUseSet.
