(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import Morphisms.

Require Export FLAM_Algebra.
Require Export Setoid.
(* end hide *)

(**
 * FLAM_Norm: Principal normalization.
 *)
(**
 * Definitions.
 *)
Inductive normalized : Prin -> Set :=
  | norm_J : forall p : Prin, normJ p -> normalized p
  | norm_conf : forall p : Prin, normJ p -> normalized p→
  | norm_integ : forall p : Prin, normJ p -> normalized p←
  | norm_decomp : forall p q : Prin, normJ p -> normJ q -> normalized (p→ ∧ q←)
with normJ : Prin -> Set :=
  | normJ_M : forall p : Prin, normM p -> normJ p
  | normJ_conj : forall p q : Prin, normJ p -> normJ q -> normJ (p ∧ q)
with normM : Prin -> Set :=
  | normM_L : forall p : Prin, normL p -> normM p
  | normM_disj : forall p q : Prin, normM p -> normM q -> normM (p ∨ q)
with normL : Prin -> Set :=
  | normL_name : forall n : name, normL (native n)
  | normL_top : normL ⊤
  | normL_bot : normL ⊥
  | normL_own : forall p q : Prin, normL p -> normL q -> normL (p:q)
.

Create HintDb norm.
Hint Resolve norm_J : norm.
Hint Resolve norm_conf : norm.
Hint Resolve norm_integ : norm.
Hint Resolve norm_decomp : norm.
Hint Resolve normJ_M : norm.
Hint Resolve normJ_conj : norm.
Hint Resolve normM_L : norm.
Hint Resolve normM_disj : norm.
Hint Resolve normL_name : norm.
Hint Resolve normL_top : norm.
Hint Resolve normL_bot : norm.
Hint Resolve normL_own : norm.

Tactic Notation "autonorm" := auto with norm.
Tactic Notation "eautonorm" := eauto with norm.

(**
 ** Normalizing confidentiality projections.
 *)
Definition normConf (p : Prin) (pfp : normalized p) : sigT normalized :=
  match pfp with
  | norm_J a pfa => existT _ a→ (norm_conf a pfa)
  | norm_conf a pfa => existT _ a→ (norm_conf a pfa)
  | norm_integ a pfa => existT _ ⊥ (norm_J _ (normJ_M _ (normM_L _ normL_bot)))
  | norm_decomp a b pfa _ => existT _ a→ (norm_conf a pfa)
  end.

Lemma normConf_correct:
  forall (p : Prin) (pf : normalized p),
    p→ == projT1 (normConf p pf).
Proof.
  intros.
  destruct pf ; try (rewrite conjConf) ; autoalg.
Qed.

(**
 ** Normalizing integrity projections.
 *)
Definition normInteg (p : Prin) (pfp : normalized p) : sigT normalized :=
  match pfp with
  | norm_J a pfa => existT _ a← (norm_integ a pfa)
  | norm_conf a pfa => existT _ ⊥ (norm_J _ (normJ_M _ (normM_L _ normL_bot)))
  | norm_integ a pfa => existT _ a← (norm_integ a pfa)
  | norm_decomp a b _   pfb => existT _ b← (norm_integ b pfb)
  end.

Lemma normInteg_correct:
  forall (p : Prin) (pf : normalized p),
    p← == projT1 (normInteg p pf).
Proof.
  intros.
  destruct pf ; try (rewrite conjInteg) ; autoalg.
Qed.

(**
 ** Normalizing conjunctions.
 *)
Definition normConj (p : Prin) (pfp : normalized p)
                    (q : Prin) (pfq : normalized q) : sigT normalized :=
  match pfp, pfq with
  | norm_J a pfa, norm_J c pfc =>
      existT _ (a ∧ c)
        (norm_J _ (normJ_conj _ _ pfa pfc))
  | norm_J a pfa, norm_conf c pfc =>
      existT _ ((a ∧ c)→ ∧ a←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) pfa)
  | norm_J a pfa, norm_integ d pfd =>
      existT _ (a→ ∧ (a ∧ d)←)
        (norm_decomp _ _ pfa (normJ_conj _ _ pfa pfd))
  | norm_J a pfa, norm_decomp c d pfc pfd =>
      existT _ ((a ∧ c)→ ∧ (a ∧ d)←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) (normJ_conj _ _ pfa pfd))

  | norm_conf a pfa, norm_J c pfc =>
      existT _ ((a ∧ c)→ ∧ c←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) pfc)
  | norm_conf a pfa, norm_conf c pfc =>
      existT _ ((a ∧ c)→)
        (norm_conf _ (normJ_conj _ _ pfa pfc))
  | norm_conf a pfa, norm_integ d pfd =>
      existT _ (a→ ∧ d←)
        (norm_decomp _ _ pfa pfd)
  | norm_conf a pfa, norm_decomp c d pfc pfd =>
      existT _ ((a ∧ c)→ ∧ d←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) pfd)

  | norm_integ b pfb, norm_J c pfc =>
      existT _ (c→ ∧ (b ∧ c)←)
        (norm_decomp _ _ pfc (normJ_conj _ _ pfb pfc))
  | norm_integ b pfb, norm_conf c pfc =>
      existT _ (c→ ∧ b←)
        (norm_decomp _ _ pfc pfb)
  | norm_integ b pfb, norm_integ d pfd =>
      existT _ ((b ∧ d)←)
        (norm_integ _ (normJ_conj _ _ pfb pfd))
  | norm_integ b pfb, norm_decomp c d pfc pfd =>
      existT _ (c→ ∧ (b ∧ d)←)
        (norm_decomp _ _ pfc (normJ_conj _ _ pfb pfd))

  | norm_decomp a b pfa pfb, norm_J c pfc =>
      existT _ ((a ∧ c)→ ∧ (b ∧ c)←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) (normJ_conj _ _ pfb pfc))
  | norm_decomp a b pfa pfb, norm_conf c pfc =>
      existT _ ((a ∧ c)→ ∧ b←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) pfb)
  | norm_decomp a b pfa pfb, norm_integ d pfd =>
      existT _ (a→ ∧ (b ∧ d)←)
        (norm_decomp _ _ pfa (normJ_conj _ _ pfb pfd))
  | norm_decomp a b pfa pfb, norm_decomp c d pfc pfd =>
      existT _ ((a ∧ c)→ ∧ (b ∧ d)←)
        (norm_decomp _ _ (normJ_conj _ _ pfa pfc) (normJ_conj _ _ pfb pfd))
  end.

Lemma normConj_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normalized q),
    p ∧ q == projT1 (normConj p pfp q pfq).
Proof.
  intros.
  destruct pfp;
      destruct pfq;
      simpl;
      try (rewrite conjConf);
      try (rewrite conjInteg);
      autoalg.
    rewrite conjPermute.
    autoalg.

    rewrite conjAssocRev.
    autoalg.

    rewrite conjPermute2in4.
    autoalg.

    rewrite conjAssoc.
    autoalg.

    rewrite conjAssocRev.
    rewrite conjPermute.
    autoalg.

    rewrite conjComm.
    rewrite conjPermute.
    autoalg.

    rewrite conjPermute2in4.
    autoalg.
Qed.

(**
 ** Normalizing disjunctions.
 *)

(**
 *** Disjunctions of meets with joins.
 *)
Fixpoint normDisjMJ (p : Prin) (pfp : normM p)
                    (q : Prin) (pfq : normJ q) : sigT normJ :=
    match pfq with
    | normJ_M a pfa =>
        existT _ (p ∨ a)
          (normJ_M _ (normM_disj _ _ pfp pfa))
    | normJ_conj a b pfa pfb =>
        let pora := (normDisjMJ p pfp a pfa) in
        let porb := (normDisjMJ p pfp b pfb) in
          existT _
            ((projT1 pora) ∧ (projT1 porb))
            (normJ_conj _ _ (projT2 pora) (projT2 porb))
    end.

Lemma normDisjMJ_correct:
  forall (p : Prin) (pfp : normM p) (q : Prin) (pfq : normJ q),
    p ∨ q == projT1 (normDisjMJ p pfp q pfq).
Proof.
  intros.
  induction pfq ; simpl ; autoalg.
    rewrite <- IHpfq1.
    rewrite <- IHpfq2.
    autoalg.
Qed.

(**
 *** Disjunctions of joins with joins.
 *)
Fixpoint normDisjJJ (p : Prin) (pfp : normJ p)
                    (q : Prin) (pfq : normJ q) : sigT normJ :=
  match pfp, pfq with
  | normJ_M a pfa, normJ_M c pfc =>
      existT _ (a ∨ c)
        (normJ_M _ (normM_disj _ _ pfa pfc))
  | normJ_M a pfa, normJ_conj c d pfc pfd =>
      let aorc := normDisjMJ a pfa c pfc in
      let aord := normDisjMJ a pfa d pfd in
        existT _ ((projT1 aorc) ∧ (projT1 aord))
          (normJ_conj _ _ (projT2 aorc) (projT2 aord))

  | normJ_conj a b pfa pfb, _ =>
      let aorq := normDisjJJ a pfa q pfq in
      let borq := normDisjJJ b pfb q pfq in
        existT _ ((projT1 aorq) ∧ (projT1 borq))
          (normJ_conj _ _ (projT2 aorq) (projT2 borq))
  end.

Lemma normDisjJJ_correct:
  forall (p : Prin) (pfp : normJ p) (q : Prin) (pfq : normJ q),
    p ∨ q == projT1 (normDisjJJ p pfp q pfq).
Proof.
  intros.
  induction pfp.
    destruct pfq ; autoalg.
    simpl.
    repeat (rewrite <- normDisjMJ_correct).
    autoalg.

    simpl.
    rewrite <- IHpfp1.
    rewrite <- IHpfp2.
    autoalg.
Qed.

(**
 *** General disjunctions.
 *)
Definition normDisj (p : Prin) (pfp : normalized p)
                    (q : Prin) (pfq : normalized q) : sigT normalized :=
  match pfp, pfq with
  | norm_J a pfa, norm_J c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc) (norm_J _ (projT2 aorc))
  | norm_J a pfa, norm_conf c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc)→ (norm_conf _ (projT2 aorc))
  | norm_J a pfa, norm_integ d pfd =>
      let aord := normDisjJJ a pfa d pfd in
        existT _ (projT1 aord)← (norm_integ _ (projT2 aord))
  | norm_J a pfa, norm_decomp c d pfc pfd =>
      let aorc := normDisjJJ a pfa c pfc in
      let aord := normDisjJJ a pfa d pfd in
        existT _ ((projT1 aorc)→ ∧ (projT1 aord)←)
          (norm_decomp _ _ (projT2 aorc) (projT2 aord))

  | norm_conf a pfa, norm_J c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc)→
          (norm_conf _ (projT2 aorc))
  | norm_conf a pfa, norm_conf c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc)→
          (norm_conf _ (projT2 aorc))
  | norm_conf _ _, norm_integ _ _ =>
      existT _ ⊥
        (norm_J _ (normJ_M _ (normM_L _ normL_bot)))
  | norm_conf a pfa, norm_decomp c _ pfc _ =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc)→
          (norm_conf _ ((projT2 aorc)))

  | norm_integ b pfb, norm_J c pfc =>
      let borc := normDisjJJ b pfb c pfc in
        existT _ (projT1 borc)←
          (norm_integ _ ((projT2 borc)))
  | norm_integ _ _, norm_conf _ _ =>
      existT _ ⊥
        (norm_J _ (normJ_M _ (normM_L _ normL_bot)))
  | norm_integ b pfb, norm_integ d pfd =>
      let bord := normDisjJJ b pfb d pfd in
        existT _ (projT1 bord)←
          (norm_integ _ ((projT2 bord)))
  | norm_integ b pfb, norm_decomp _ d _ pfd =>
      let bord := normDisjJJ b pfb d pfd in
        existT _ (projT1 bord)←
          (norm_integ _ ((projT2 bord)))

  | norm_decomp a b pfa pfb, norm_J c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
      let borc := normDisjJJ b pfb c pfc in
        existT _ ((projT1 aorc)→ ∧ (projT1 borc)←)
          (norm_decomp _ _ (projT2 aorc) (projT2 borc))
  | norm_decomp a _ pfa _, norm_conf c pfc =>
      let aorc := normDisjJJ a pfa c pfc in
        existT _ (projT1 aorc)→
          (norm_conf _ (projT2 aorc))
  | norm_decomp _ b _ pfb, norm_integ d pfd =>
      let bord := normDisjJJ b pfb d pfd in
        existT _ (projT1 bord)←
          (norm_integ _ ((projT2 bord)))
  | norm_decomp a b pfa pfb, norm_decomp c d pfc pfd =>
      let aorc := normDisjJJ a pfa c pfc in
      let bord := normDisjJJ b pfb d pfd in
        existT _ ((projT1 aorc)→ ∧ (projT1 bord)←)
          (norm_decomp _ _ (projT2 aorc) (projT2 bord))
  end.

Lemma normDisj_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normalized q),
    p ∨ q == projT1 (normDisj p pfp q pfq).
Proof.
  intros.
  destruct pfp;
      destruct pfq;
      simpl;
      repeat (rewrite <- normDisjJJ_correct);
      try (rewrite disjConf);
      try (rewrite disjInteg);
      autoalg.

    rewrite disjLDistConj.
    rewrite disjConfR.
    rewrite disjIntegR.
    autoalg.

    rewrite disjLDistConj.
    autoalg.

    rewrite disjLDistConj.
    autoalg.

    rewrite disjRDistConj.
    rewrite disjConfL.
    rewrite disjIntegL.
    autoalg.

    rewrite disjRDistConj.
    autoalg.

    rewrite disjRDistConj.
    autoalg.

    rewrite disjRDistConj.
    repeat (rewrite disjLDistConj).
    autoalg.
Qed.

(**
 ** Normalizing ownership projections.
 *)

(**
 *** Meet owns literal.
 *)
Fixpoint normOwnsML (p : Prin) (pfp : normM p)
                    (q : Prin) (pfq : normL q) : sigT normM :=
  match pfp with
  | normM_L a pfa =>
      existT _ (a:q)
        (normM_L _ (normL_own _ _ pfa pfq))
  | normM_disj a b pfa pfb =>
      let aq := normOwnsML a pfa q pfq in
      let bq := normOwnsML b pfb q pfq in
        existT _ ((projT1 aq) ∨ (projT1 bq))
          (normM_disj _ _ (projT2 aq) (projT2 bq))
  end.

Lemma normOwnsML_correct:
  forall (p : Prin) (pfp : normM p) (q : Prin) (pfq : normL q),
    p:q == projT1 (normOwnsML p pfp q pfq).
Proof.
  intros.
  induction pfp ; autoalg.
  simpl.
  rewrite <- IHpfp1.
  rewrite <- IHpfp2.
  autoalg.
Qed.

(**
 *** Join owns literal.
 *)
Fixpoint normOwnsJL (p : Prin) (pfp : normJ p)
                    (q : Prin) (pfq : normL q) : sigT normJ :=
  match pfp with
  | normJ_M a pfa =>
      let aq := normOwnsML a pfa q pfq in
        existT _ (projT1 aq)
          (normJ_M _ (projT2 aq))
  | normJ_conj a b pfa pfb =>
      let aq := normOwnsJL a pfa q pfq in
      let bq := normOwnsJL b pfb q pfq in
        existT _ ((projT1 aq) ∧ (projT1 bq))
          (normJ_conj _ _ (projT2 aq) (projT2 bq))
  end.

Lemma normOwnsJL_correct:
  forall (p : Prin) (pfp : normJ p) (q : Prin) (pfq : normL q),
    p:q == projT1 (normOwnsJL p pfp q pfq).
Proof.
  intros.
  induction pfp ; simpl ; auto using normOwnsML_correct with prinalg.
  rewrite <- IHpfp1.
  rewrite <- IHpfp2.
  autoalg.
Qed.

(**
 *** Join owns meet.
 *)
Fixpoint normOwnsJM (p : Prin) (pfp : normJ p)
                    (q : Prin) (pfq : normM q) : sigT normJ :=
  match pfq with
  | normM_L c pfc => normOwnsJL p pfp c pfc
  | normM_disj c d pfc pfd =>
      let pc := normOwnsJM p pfp c pfc in
      let pd := normOwnsJM p pfp d pfd in
        normDisjJJ (projT1 pc) (projT2 pc) (projT1 pd) (projT2 pd)
  end.

Lemma normOwnsJM_correct:
  forall (p : Prin) (pfp : normJ p) (q : Prin) (pfq : normM q),
    p:q == projT1 (normOwnsJM p pfp q pfq).
Proof.
  intros.
  induction pfq ; simpl ; auto using normOwnsJL_correct with prinalg.
  rewrite disjDistOwn.
  rewrite IHpfq1.
  rewrite IHpfq2.
  destruct (normOwnsJM p pfp p0 pfq1).
  destruct (normOwnsJM p pfp q pfq2).
  simpl.
  apply normDisjJJ_correct.
Qed.

(**
 *** Join owns join.
 *)
Fixpoint normOwnsJJ (p : Prin) (pfp : normJ p)
                    (q : Prin) (pfq : normJ q) : sigT normJ :=
  match pfq with
  | normJ_M c pfc => normOwnsJM p pfp c pfc
  | normJ_conj c d pfc pfd =>
      let pc := normOwnsJJ p pfp c pfc in
      let pd := normOwnsJJ p pfp d pfd in
        existT _ ((projT1 pc) ∧ (projT1 pd))
          (normJ_conj (projT1 pc) (projT1 pd) (projT2 pc) (projT2 pd))
  end.

Lemma normOwnsJJ_correct:
  forall (p : Prin) (pfp : normJ p) (q : Prin) (pfq : normJ q),
    p:q == projT1 (normOwnsJJ p pfp q pfq).
Proof.
  intros.
  induction pfq ; simpl ; auto using normOwnsJM_correct with prinalg.
  rewrite <- IHpfq1, <- IHpfq2.
  autoalg.
Qed.

(**
 *** Principal owns literal.
 *)
Definition normOwnsPL (p : Prin) (pfp : normalized p)
                      (q : Prin) (pfq : normL q) : sigT normalized :=
  match pfp with
  | norm_J a pfa =>
      let aq := (normOwnsJL a pfa q pfq) in
        existT _ (projT1 aq) (norm_J _ (projT2 aq))
  | norm_conf a pfa =>
      let aq := normOwnsJL a pfa q pfq in
        existT _ ((projT1 aq)→)
          (norm_conf _ (projT2 aq))
  | norm_integ b pfb =>
      let bq := normOwnsJL b pfb q pfq in
        existT _ ((projT1 bq)←)
          (norm_integ _ (projT2 bq))
  | norm_decomp a b pfa pfb =>
      let aq := normOwnsJL a pfa q pfq in
      let bq := normOwnsJL b pfb q pfq in
        existT _ ((projT1 aq)→ ∧ (projT1 bq)←)
          (norm_decomp (projT1 aq) (projT1 bq) (projT2 aq) (projT2 bq))
  end.

Lemma normOwnsPL_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normL q),
    p:q == projT1 (normOwnsPL p pfp q pfq).
Proof.
  intros.
  destruct pfp;
      simpl;
      repeat rewrite <- normOwnsJL_correct;
      auto using normOwnsJL_correct with prinalg.
    rewrite ownDistConj.
    autoalg.
Qed.

(**
 *** Principal owns meet.
 *)
Fixpoint normOwnsPM (p : Prin) (pfp : normalized p)
                    (q : Prin) (pfq : normM q) : sigT normalized :=
  match pfq with
  | normM_L c pfc => normOwnsPL p pfp c pfc
  | normM_disj c d pfc pfd =>
      let pc := normOwnsPM p pfp c pfc in
      let pd := normOwnsPM p pfp d pfd in
        normDisj (projT1 pc) (projT2 pc) (projT1 pd) (projT2 pd)
  end.

Lemma normOwnsPM_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normM q),
    p:q == projT1 (normOwnsPM p pfp q pfq).
Proof.
  intros.
  induction pfq ; simpl ; auto using normOwnsPL_correct with prinalg.
  rewrite disjDistOwn.
  rewrite IHpfq1.
  rewrite IHpfq2.
  destruct (normOwnsPM p pfp p0 pfq1).
  destruct (normOwnsPM p pfp q pfq2).
  simpl.
  apply normDisj_correct.
Qed.

(**
 *** Principal owns join.
 *)
Fixpoint normOwnsPJ (p : Prin) (pfp : normalized p)
                    (q : Prin) (pfq : normJ q) : sigT normalized :=
  match pfq with
  | normJ_M c pfc => normOwnsPM p pfp c pfc
  | normJ_conj c d pfc pfd =>
      let pc := normOwnsPJ p pfp c pfc in
      let pd := normOwnsPJ p pfp d pfd in
        normConj (projT1 pc) (projT2 pc) (projT1 pd) (projT2 pd)
  end.

Lemma normOwnsPJ_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normJ q),
    p:q == projT1 (normOwnsPJ p pfp q pfq).
Proof.
  intros.
  induction pfq ; simpl ; auto using normOwnsPM_correct with prinalg.
  rewrite <- normConj_correct, <- IHpfq1, <- IHpfq2.
  autoalg.
Qed.

(**
 *** General ownership projections.
 *)
Definition normOwns (p : Prin) (pfp : normalized p)
                    (q : Prin) (pfq : normalized q) : sigT normalized :=
  match pfq with
  | norm_J c pfc =>
      normOwnsPJ p pfp c pfc
  | norm_conf c pfc =>
      let pc := normOwnsPJ p pfp c pfc in
        normConf (projT1 pc) (projT2 pc)
  | norm_integ d pfd =>
      let pd := normOwnsPJ p pfp d pfd in
        normInteg (projT1 pd) (projT2 pd)
  | norm_decomp c d pfc pfd =>
      let pc := normOwnsPJ p pfp c pfc in
      let pd := normOwnsPJ p pfp d pfd in
      let pc_conf := normConf (projT1 pc) (projT2 pc) in
      let pd_integ := normInteg (projT1 pd) (projT2 pd) in
        normConj (projT1 pc_conf) (projT2 pc_conf)
          (projT1 pd_integ) (projT2 pd_integ)
  end.

Lemma normOwns_correct:
  forall (p : Prin) (pfp : normalized p) (q : Prin) (pfq : normalized q),
    p:q == projT1 (normOwns p pfp q pfq).
Proof.
  intros.
  destruct pfq;
      simpl;
      try (rewrite <- normConj_correct);
      try (rewrite <- normConf_correct);
      try (rewrite <- normInteg_correct);
      repeat rewrite <- normOwnsPJ_correct;
      autoalg.

    rewrite ownConfAssoc, ownIntegAssoc.
    autoalg.
Qed.

(**
 ** General principal normalization.
 *)
Fixpoint norm p : sigT normalized :=
  match p with
  | ⊤ =>
      existT _ ⊤
        (norm_J _ (normJ_M _ (normM_L _ normL_top)))
  | ⊥ =>
      existT _ ⊥
        (norm_J _ (normJ_M _ (normM_L _ normL_bot)))
  | native n =>
      existT _ (native n)
        (norm_J _ (normJ_M _ (normM_L _ (normL_name n))))
  | a ∧ b =>
      let na := norm a in
      let nb := norm b in
        normConj (projT1 na) (projT2 na) (projT1 nb) (projT2 nb)
  | a ∨ b =>
      let na := norm a in
      let nb := norm b in
        normDisj (projT1 na) (projT2 na) (projT1 nb) (projT2 nb)
  | a→ =>
      let na := norm a in
        normConf (projT1 na) (projT2 na)
  | a← =>
      let na := norm a in
        normInteg (projT1 na) (projT2 na)
  | a:b =>
      let na := norm a in
      let nb := norm b in
        normOwns (projT1 na) (projT2 na) (projT1 nb) (projT2 nb)
  end.

Lemma norm_correct:
  forall p, p == projT1 (norm p).
Proof.
  intros;
    induction p;
    simpl;
    try (rewrite IHp at 1);
    try (rewrite IHp1 at 1);
    try (rewrite IHp2 at 1);
    auto using normConj_correct, normDisj_correct, normConf_correct,
      normInteg_correct, normOwns_correct with prinalg.
Qed.

(**
 * Normalization properties.
 *)

Scheme normJ_ind' := Induction for normJ Sort Prop
with normM_ind' := Induction for normM Sort Prop
with normL_ind' := Induction for normL Sort Prop.

Combined Scheme norm_mutind from normJ_ind', normM_ind', normL_ind'.

(**
 ** Idempotence.
 *)
(* The master lemma for the idempotence of normalization. *)
Lemma normIdemp_master:
  (forall p (pfp : normJ p),
    exists (pfp' : normJ p),
      norm p = existT _ p (norm_J _ pfp'))
  /\ (forall p (pfp : normM p),
       exists (pfp' : normM p),
         norm p = existT _ p (norm_J _ (normJ_M _ pfp')))
  /\ (forall p (pfp : normL p),
       exists (pfp' : normL p),
         norm p = existT _ p (norm_J _ (normJ_M _ (normM_L _ pfp')))).
Proof.
  apply norm_mutind ; intros ; simpl ; eauto.
    destruct H.
    rewrite H.
    eauto.

    destruct H.
    destruct H0.
    rewrite H.
    rewrite H0.
    simpl.
    eauto.

    destruct H.
    eauto.

    destruct H.
    destruct H0.
    rewrite H.
    rewrite H0.
    simpl.
    eauto.

    destruct H0.
    rewrite H0.
    destruct H.
    rewrite H.
    simpl.
    eauto.
Qed.

Lemma normIdemp_J:
  forall p (pfp : normJ p),
    exists (pfp' : normJ p),
      norm p = existT _ p (norm_J _ pfp').
Proof.
  apply normIdemp_master.
Qed.

Lemma normIdemp:
  forall p (pfp : normJ p) q (pfq : normJ q),
    exists (pfp' : normJ p) (pfq' : normJ q),
      norm (p→ ∧ q←) =
        existT _ (p→ ∧ q←)
          (norm_decomp _ _ pfp' pfq').
Proof.
  intros.
  simpl.
  pose proof (normIdemp_J p pfp).
  pose proof (normIdemp_J q pfq).
  destruct H.
  destruct H0.
  rewrite H.
  rewrite H0.
  simpl. 
  eauto.
Qed.

(**
 ** Structure of normalized principals.
 *)
Lemma normForm:
  forall p, exists a b (pfa : normJ a) (pfb : normJ b), p == a→ ∧ b←.
Proof.
  intro.
  pose proof (norm_correct p).
  destruct (norm p).
  destruct n ; simpl ; eautoalg ; simpl in H.
    exists p0, p0, n, n.
    autoalg.

    exists p0, ⊥, n.
    exists (normJ_M _ (normM_L _ normL_bot)).
    autoalg.

    exists ⊥, p0.
    exists (normJ_M _ (normM_L _ normL_bot)), n.
    eautoalg.
Qed.

Lemma normJorJ:
  forall (p : Prin) (pfp : normJ p) (q : Prin) (pfq : normJ q),
    exists r (pfr : normJ r), norm (p ∨ q) = existT _ r (norm_J _ pfr).
Proof.
  intros.
  pose proof normIdemp_J p pfp.
  pose proof normIdemp_J q pfq.
  destruct H.
  destruct H0.
  simpl.
  rewrite H.
  rewrite H0.
  simpl.
  eauto.
Qed.

(**
 ** Equivalence of normalized principals and the setoid compatibility of
 ** normalization
 *)
Definition normEq (np nq : sigT normalized) :=
  projT1 np == projT1 nq.

Infix "=n=" := normEq (at level 70, no associativity).

Lemma normEq_refl : Reflexive normEq.
Proof.
  intro.
  destruct x.
  unfold normEq.
  simpl.
  autoalg.
Qed.

Lemma normEq_symm : Symmetric normEq.
Proof.
  intro. intros.
  destruct x, y.
  destruct x, x0 ; unfold normEq ; simpl ; autoalg.
Qed.

Lemma normEq_trans : Transitive normEq.
Proof.
  intro. intros.
  destruct x, y, z.
  destruct x, x0, x1 ; simpl in * ; eapply trans ; eauto ; autoalg.
Qed.

Instance normSetoid : Equivalence normEq.
Proof.
  split ; [eapply normEq_refl | eapply normEq_symm | eapply normEq_trans].
Qed.

Instance normProper : Proper (prinEq ==> normEq) norm.
Proof.
  intro. intros.
  pose proof norm_correct x.
  pose proof norm_correct y.
  destruct (norm x), (norm y).
  unfold normEq.
  simpl in *.
  apply trans with (q := x) ; autoalg.
  apply trans with (q := y) ; autoalg.
Qed.

(**
 ** Miscellaneous properties
 *)

Axiom normJConfEq:
  forall p q,
    normJ p -> normJ q -> p→ == q→ -> p == q.

Axiom normJIntegEq:
  forall p q,
    normJ p -> normJ q -> p← == q← -> p == q.

Lemma normJConfIsBotImpliesBot:
  forall p,
    normJ p -> p→ == ⊥ -> p == ⊥.
Proof.
  intros.
  rewrite <- botConf in H0.
  apply normJConfEq ; autonorm.
Qed.

Lemma normJIntegIsBotImpliesBot:
  forall p,
    normJ p -> p← == ⊥ -> p == ⊥.
Proof.
  intros.
  rewrite <- botInteg in H0.
  apply normJIntegEq ; autonorm.
Qed.

Lemma normJIsConfImpliesBot:
  forall p q, normJ p -> p == q→ -> p == ⊥.
Proof.
  intros.
  apply normJIntegIsBotImpliesBot ; autonorm.
  rewrite <- (confInteg q), H0. autoalg.
Qed.

Lemma normJIsIntegImpliesBot:
  forall p q, normJ p -> p == q← -> p == ⊥.
Proof.
  intros.
  apply normJConfIsBotImpliesBot ; autonorm.
  rewrite <- (integConf q), H0. autoalg.
Qed.

Lemma normJDecomp:
  forall p q r,
    normJ p -> normJ q -> normJ r -> p == q→ ∧ r← -> p == q /\ p == r.
Proof.
  intros.
  split.
  apply normJConfEq with (q := q) ; auto.
  setoid_replace (q→) with ((q→ ∧ r←)→).
    apply confSubst. auto.
    rewrite conjConf. autoalg.

  apply normJIntegEq with (q := r) ; auto.
  setoid_replace (r←) with ((q→ ∧ r←)←).
    apply integSubst. auto.
    rewrite conjInteg. autoalg.
Qed.

Lemma confIsIntegImpliesBot:
  forall p q, p→ == q← -> p→ == ⊥ /\ q← == ⊥.
Proof.
  intros.
  split.
    setoid_replace ⊥ with (q←→) by autoalg.
    setoid_replace (p→) with (p→→) by autoalg.
    apply confSubst.
    auto.

    setoid_replace ⊥ with (p→←) by autoalg.
    setoid_replace (q←) with (q←←) by autoalg.
    apply integSubst.
    autoalg.
Qed.
