(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Export FLAM_Algebra.
Require Export FLAM_DelSet.
Require Import Morphisms.
Require Export Setoid.
(* end hide *)

(**
 * Principal hierarchy
 *)
(**
 ** Definitions
 *)

(*
 * A principal hierarchy maps names of native principals to sets of delegations.
 *)
Module PH.
  Definition t := name -> DelSet.t.
  Definition empty := fun _ : name => DelSet.empty.

  Definition phEq (H1 H2 : t) :=
    forall n, DelSet.sameSet (H1 n) (H2 n).

  Notation "P1 =H= P2" := (phEq P1 P2) (at level 70, no associativity).

  Instance phEqSetoid : Equivalence phEq.
  Proof.
    destruct DelSet.sameSetSetoid.

    split.
    { intro. intros. intro. auto. }
    { intro. intros. intro. pose proof H n. auto. }
    { intro. intros. intro. pose proof H n. pose proof H0 n. eauto. }
  Qed.

  Lemma phEqRefl : Reflexive phEq.
  Proof.
    apply phEqSetoid.
  Qed.

  Hint Resolve phEqRefl.
  Hint Resolve phEqRefl : flam.

  Lemma phEqTrans : Transitive phEq.
  Proof.
    apply phEqSetoid.
  Qed.

  Hint Resolve phEqTrans : flam.

  (* Representation invariant. *)
  Definition repOk (H : t) :=
    forall cn d, DelSet.In d (H cn) -> DelT.hostedAt cn d.

  Instance repOkProper:
    Proper (phEq ==> iff) repOk.
  Proof.
    intro. intros.
    split ; intros.
    { intro. intros. apply H0. rewrite (H cn). auto. }
    { intro. intros. apply H0. rewrite <- (H cn). auto. }
  Qed.

  Lemma emptyOk: repOk empty.
  Proof.
    intro. intros.
    destruct d. unfold empty in H.
    contradict H.
    apply DelSet.emptyIsEmpty.
  Qed.

  (* Principal hierarchy with a single element. *)
  Definition singleton (d : Del.t) :=
    fun pName =>
      match d with
      | DelT.make _ _ _ rName =>
          if name_dec pName rName then DelSet.singleton d else DelSet.empty
      end.

  Lemma singletonOk: forall d, repOk (singleton d).
  Proof.
    intros.
    destruct d.
    intro. intros.
    destruct d.
    simpl.
    unfold singleton in H.
    destruct (name_dec cn n).
    {
      subst.
      pose proof DelSet.singletonInv _ _ H.
      inversion H0. auto.
    }
    { contradict H. apply DelSet.emptyIsEmpty. }
  Qed.

  (* Union of two principal hierarchies. *)
  Definition union (H H' : t) : t :=
    fun pName => DelSet.union (H pName) (H' pName).

  Instance unionProper:
    Proper (phEq ==> phEq ==> phEq) union.
  Proof.
    intro. intros. intro. intros.
    intro.
    unfold union.
    pose proof H n.
    pose proof H0 n.
    rewrite H1, H2.
    auto with flam.
  Qed.

  Lemma unionOk: forall H1 H2, repOk H1 -> repOk H2 -> repOk (union H1 H2).
  Proof.
    intros. intro. intros.
    destruct (DelSet.unionInv _ _ _ H3).
    { apply H. auto. }
    { apply H0. auto. }
  Qed.
  Lemma okUnion: forall H1 H2, repOk (union H1 H2) -> repOk H1 /\ repOk H2.
  Proof.
    intros.
    split.
    {
      intro. intros.
      unfold repOk in H.
      apply H with (d := d) (cn := cn).
      eapply DelSet.unionIncreasesL.
      eauto.
    }
    {
      intro. intros.
      unfold repOk in H.
      apply H with (d := d) (cn := cn).
      eapply DelSet.unionIncreasesR.
      eauto.
    }
  Qed.

  (* Containment relation. *)
  Definition included (H H' : t) :=
    forall p, DelSet.included (H p) (H' p).

  Instance includedProper:
    Proper (phEq ==> phEq ==> iff) included.
  Proof.
    intro. intros. intro. intros.
    split ; intros ; intro.
    {
      pose proof H p.
      pose proof H0 p.
      pose proof H1 p.
      rewrite H2, H3 in *.
      auto.
    }
    {
      pose proof H p.
      pose proof H0 p.
      pose proof H1 p.
      rewrite H2, H3 in *.
      auto.
    }
  Qed.
  
  Definition In (d : Del.t) (H : t) :=
    match d with
    | DelT.make p q l rn =>
        DelSet.In d (H rn)
    end.

  Lemma In_dec:
    forall d H, {In d H} + {~ In d H}.
  Proof.
    intros.
    destruct d.
    simpl.
    apply DelSet.In_dec.
  Qed.

  Instance InProper:
    Proper (eq ==> phEq ==> iff) In.
  Proof.
    intro. intros. intro. intros.
    split ; intros ; subst ; destruct y ; simpl in *.
    {
      pose proof H0 n.
      rewrite <- H. auto.
    }
    {
      pose proof H0 n.
      rewrite H. auto.
    }
  Qed.

  (* An alternative, members-based definition of equivalence. *)
  Definition membersEq (H1 H2 : t) :=
    forall (d : DelT.t), In d H1 <-> In d H2.

  Instance membersEqProper:
    Proper (phEq ==> phEq ==> iff) membersEq.
  Proof.
    intro. intros. intro. intros.
    unfold membersEq. setoid_rewrite H. setoid_rewrite H0.
    split ; intro ; auto.
  Qed.

  (**
   ** Properties
   *)
  Lemma eqIffMembersEq:
    forall H1 H2, repOk H1 -> repOk H2 -> (H1 =H= H2 <-> membersEq H1 H2).
  Proof.
    intros ; split ; intros ; intro ; intros.
    { destruct d. apply H3. }
    {
      intro. destruct x as [p q l rn]. split ; intro Hin.
      {
        pose proof H _ _ Hin as Hrewrite. simpl in Hrewrite. subst.
        fold (In (DelT.make p q l rn) H1) in Hin.
        fold (In (DelT.make p q l rn) H2).
        remember (DelT.make p q l rn) as d.
        destruct (H3 d).
        auto.
      }
      {
        pose proof H0 _ _ Hin as Hrewrite. simpl in Hrewrite. subst.
        fold (In (DelT.make p q l rn) H2) in Hin.
        fold (In (DelT.make p q l rn) H1).
        remember (DelT.make p q l rn) as d.
        destruct (H3 d).
        auto.
      }
    }
  Qed.

  Lemma inclusionInv:
    forall d H' H, included H' H -> In d H' -> In d H.
  Proof.
    intros.
    destruct d.
    eapply DelSet.inclusionInv ; eauto.
  Qed.

  Lemma unionInv:
    forall d H1 H2,
      In d (union H1 H2) -> In d H1 \/ In d H2.
  Proof.
    intros.
    destruct d.
    eapply DelSet.unionInv ; eauto.
  Qed.

  Lemma unionInv3:
    forall d H1 H2 H3,
      In d (union H1 (union H2 H3))
        -> In d H1 \/ In d H2 \/ In d H3.
  Proof.
    intros.
    assert (In d H1 \/ In d (union H2 H3)) as Hinv by (eapply unionInv ; auto).
    destruct Hinv ; auto.
    right.
    eapply unionInv ; auto.
  Qed.

  Lemma unionIntro:
    forall d H1 H2,
      In d H1 \/ In d H2 -> In d (union H1 H2).
  Proof.
    intros.
    destruct H, d ; apply DelSet.unionIntro ; auto.
  Qed.

  Lemma emptyIsEmpty : forall d, ~ In d empty.
  Proof.
    intros.
    destruct d.
    simpl.
    unfold empty.
    apply DelSet.emptyIsEmpty.
  Qed.

  Lemma emptyZero : forall H, union empty H =H= H.
  Proof.
    intros.
    intro.
    unfold union.
    unfold empty.
    apply DelSet.emptyZero.
  Qed.

  Hint Resolve emptyZero : flam.

  Lemma unionEmptyEmpty : union empty empty =H= empty.
  Proof.
    rewrite emptyZero ; auto with flam.
  Qed.

  Hint Resolve unionEmptyEmpty : flam.

  Lemma lessThanEmpty:
    forall H, included H empty -> H =H= empty.
  Proof.
    intros.
    intro.
    pose proof H0 n.
    eapply DelSet.lessThanEmpty ; auto.
  Qed.

  (* Union is commutative. *)
  Lemma unionComm:
    forall H H', union H H' =H= union H' H.
  Proof.
    intros.
    intro.
    apply DelSet.unionComm.
  Qed.

  (* Union is associative. *)
  Lemma unionAssoc:
    forall H1 H2 H3,
      union (union H1 H2) H3 =H= union H1 (union H2 H3).
  Proof.
    intros.
    intro.
    unfold union.
    apply DelSet.unionAssoc.
  Qed.

  Hint Resolve unionAssoc : flam.

  Lemma unionIdemp:
    forall H, H =H= union H H.
  Proof.
    intros.
    intro.
    unfold union.
    symmetry.
    apply DelSet.unionIdemp.
  Qed.

  Lemma unionAbsorbs:
    forall H1 H2, included H2 H1 -> union H1 H2 =H= H1.
  Proof.
    intros.
    intro.
    unfold union.
    pose proof H n.
    apply DelSet.unionAbsorbs.
    auto.
  Qed.

  Lemma emptyMinimal:
    forall H, included empty H.
  Proof.
    intros.
    intro.
    apply DelSet.emptyMinimal.
  Qed.

  Hint Resolve emptyMinimal : flam.

  Lemma unionMinimal:
    forall H1 H2 H,
      included H1 H -> included H2 H -> included (union H1 H2) H.
  Proof.
    intros.
    intro.
    pose proof H0 p.
    pose proof H3 p.
    apply DelSet.unionMinimal ; auto.
  Qed.

  Hint Resolve unionMinimal : flam.

  Lemma inclusionRefl : Reflexive included.
  Proof.
    intro.
    intro.
    apply DelSet.inclusionRefl.
  Qed.

  Hint Resolve inclusionRefl : flam.

  Lemma inclusionTrans : Transitive included.
  Proof.
    intro.
    intros.
    intro.
    pose proof H p.
    pose proof H0 p.
    eapply DelSet.inclusionTrans ; eauto.
  Qed.

  Hint Resolve inclusionTrans : flam.

  Lemma unionIncreasesL:
    forall H1 H2,
      included H1 (union H1 H2).
  Proof.
    intros.
    intro.
    apply DelSet.unionIncreasesL.
  Qed.

  Hint Resolve unionIncreasesL : flam.

  Lemma unionIncreasesR:
    forall H1 H2,
      included H2 (union H1 H2).
  Proof.
    intros.
    rewrite unionComm.
    apply unionIncreasesL.
  Qed.

  Hint Resolve unionIncreasesR : flam.

  (* If a union is empty, then its components are empty too. *)
  Lemma emptyUnion:
    forall (H H' : t), empty =H= union H H' -> H =H= empty.
  Proof.
    intros.
    intro.
    unfold empty.
    pose proof H0 n.
    unfold union in H1.
    pose proof DelSet.unionIncreasesL (H n) (H' n).
    rewrite <- H1 in H2.
    pose proof DelSet.lessThanEmpty (H n) H2.
    auto.
  Qed.

  Lemma emptyUnion2:
    forall (H1 H2 : t), empty =H= union H1 H2 -> H1 =H= empty /\ H2 =H= empty.
  Proof.
    intros.
    split ; [|rewrite unionComm in H] ; eapply emptyUnion ; eauto.
  Qed.

  Lemma emptyUnion3:
    forall (H1 H2 H3 : t), empty =H= union H1 (union H2 H3)
      -> H1 =H= empty /\ H2 =H= empty /\ H3 =H= empty.
  Proof.
    intros.
    pose proof emptyUnion2 _ _ H.
    destruct H0.
    symmetry in H4.
    pose proof emptyUnion2 _ _ H4.
    auto.
  Qed.

  Lemma emptyUnion4:
    forall (H1 H2 H3 H4 : t), empty =H= union H1 (union H2 (union H3 H4))
      -> H1 =H= empty /\ H2 =H= empty /\ H3 =H= empty /\ H4 =H= empty.
  Proof.
    intros.
    pose proof emptyUnion3 _ _ _ H.
    destruct H0.
    destruct H5.
    symmetry in H6.
    pose proof emptyUnion2 _ _ H6.
    auto.
  Qed.

  Lemma singletonNotEmpty:
    forall d, ~ singleton d =H= empty.
  Proof.
    intros.
    intro.
    destruct d.
    pose proof H n.
    unfold empty in H0.
    simpl in H0.
    destruct (name_dec n n) ; auto.
    eapply DelSet.singletonNotEmpty.
    eauto.
  Qed.

  Hint Resolve singletonNotEmpty : flam.

  Lemma emptyNotSingleton:
    forall d, ~ empty =H= singleton d.
  Proof.
    intros.
    intro.
    symmetry in H.
    contradict H.
    auto with flam.
  Qed.

  Hint Resolve emptyNotSingleton : flam.

  Lemma singletonInv:
    forall d d', In d (singleton d') -> d = d'.
  Proof.
    intros.
    destruct d, d'.
    unfold In in H.
    unfold singleton in H.
    destruct (name_dec n n0).
      eapply DelSet.singletonInv.
      auto.

      contradict H.
      apply DelSet.emptyIsEmpty.
  Qed.

  Hint Resolve singletonInv : flam.

  Lemma singletonIntro:
    forall d d', d = d' -> In d (singleton d').
  Proof.
    intros.
    destruct d, d'.
    unfold In.
    unfold singleton.
    destruct (name_dec n n0).
      eapply DelSet.singletonIntro.
      auto.

      inversion H.
      tauto.
  Qed.

  Hint Resolve singletonIntro : flam.

  Lemma inclusion2in3:
    forall H1 H2 H3,
      included H2 (union H1 (union H2 H3)).
  Proof.
    intros.
    apply inclusionTrans with (y := union H2 H3).
      apply unionIncreasesL.
      apply unionIncreasesR.
  Qed.

  Hint Resolve inclusion2in3 : flam.

  Lemma inclusion3in3:
    forall H1 H2 H3,
      included H3 (union H1 (union H2 H3)).
  Proof.
    intros.
    rewrite (unionComm H2 H3).
    apply inclusion2in3.
  Qed.

  Hint Resolve inclusion3in3 : flam.

  Lemma inclusion2in4:
    forall H1 H2 H3 H4,
      included H2 (union H1 (union H2 (union H3 H4))).
  Proof.
    intros.
    apply inclusionTrans with (y := union H2 (union H3 H4)).
      apply unionIncreasesL.
      apply unionIncreasesR.
  Qed.

  Hint Resolve inclusion2in4 : flam.

  Lemma inclusion3in4:
    forall H1 H2 H3 H4,
      included H3 (union H1 (union H2 (union H3 H4))).
  Proof.
    intros.
    rewrite <- (unionAssoc H2 H3 H4).
    rewrite (unionComm H2 H3).
    rewrite (unionAssoc H3 H2 H4).
    apply inclusion2in4.
  Qed.

  Hint Resolve inclusion3in4 : flam.

  Lemma inclusion4in4:
    forall H1 H2 H3 H4,
      included H4 (union H1 (union H2 (union H3 H4))).
  Proof.
    intros.
    rewrite (unionComm H3 H4).
    apply inclusion3in4.
  Qed.

  Hint Resolve inclusion4in4 : flam.
End PH.
