(* begin hide *)
(* 45678901234567890123456789012345678901234567890123456789012345678901234567 *)
Require Import Morphisms.
Require Import MSetProperties.
Require Import MSetWeakList.

Require Export Setoid.
(* end hide *)

(*
 * Mostly a backward-compatibility layer for clients of an older version of
 * FiniteSet.
 *)
Module FiniteSet (T0 : MiniDecidableType).
  Module T := Make_UDT T0.
  Module MSet := MSetWeakList.Make T.
  Import MSet.

  Module WProperties := WPropertiesOn T MSet.
  Import WProperties.

  Definition t := t.

  Definition In (x : T.t) (A : t) :=
    MSet.In x A.

  (**
   * Inclusion and equivalence
   *)

  Definition included (A B : t) := forall x, In x A -> In x B.
  Definition sameSet (A B : t) := forall x, In x A <-> In x B.

  Notation "A [=] B" := (sameSet A B) (at level 70, no associativity).

  (**
   ** Properties
   *)

  Instance sameSetSetoid : Equivalence sameSet.
  Proof.
    split.
    { intro. apply equal_refl. }
    { intro. intros. apply equal_sym. auto. }
    { intro. intros. eapply equal_trans ; eauto. }
  Qed.

  Lemma sameSetRefl : Reflexive sameSet.
  Proof.
    apply sameSetSetoid.
  Qed.

  Hint Resolve sameSetRefl.
  Hint Resolve sameSetRefl : flam.

  Lemma sameSetSymm : Symmetric sameSet.
  Proof.
    apply sameSetSetoid.
  Qed.

  Lemma sameSetTrans : Transitive sameSet.
  Proof.
    apply sameSetSetoid.
  Qed.

  Lemma inclusionRefl : Reflexive included.
  Proof.
    intro. eapply subset_refl.
  Qed.

  Hint Resolve inclusionRefl : flam.

  Lemma inclusionTrans : Transitive included.
  Proof.
    intro. intros. eapply subset_trans ; eauto.
  Qed.

  Hint Resolve inclusionTrans : flam.

  Lemma inclusionIntro:
    forall (A B : t),
      (forall x, In x A -> In x B) -> included A B.
  Proof.
    intros.
    unfold included, In in *.
    destruct A, B.
    auto.
  Qed.

  Hint Resolve inclusionIntro : flam.

  Lemma inclusionInv:
    forall (A B : t),
      included A B -> forall x, In x A -> In x B.
  Proof.
    intros.
    unfold included, In in *.
    destruct A, B.
    auto.
  Qed.

  Hint Resolve inclusionInv : flam.

  Lemma In_dec:
    forall x A, {In x A} + {~ In x A}.
  Proof.
    intros. apply In_dec.
  Qed.

  Instance InProper:
    Proper (Logic.eq ==> sameSet ==> iff) In.
  Proof.
    intro. intros. intro. intros.
    split ; intros ; subst ; apply H0 ; auto.
  Qed.

  Instance inclusionProper:
    Proper (sameSet ==> sameSet ==> iff) included.
  Proof.
    intro. intros.
    intro. intros.
    split ; intros.
    {
      intro. intros.
      apply H0.
      apply H1.
      apply H.
      auto.
    }
    {
      intro. intros.
      apply H0.
      apply H1.
      apply H.
      auto.
    }
  Qed.

  (**
   * Adding elements
   *)

  Definition add (x : T.t) (A B : t) :=
    forall x', In x' B <-> (x = x' \/ In x' A).

  (**
   ** Properties
   *)

  Lemma addInclude:
    forall A B x, add x A B -> included A B.
  Proof.
    intros.
    intro. intros.
    destruct (H x0).
    apply H2. auto.
  Qed.

  (**
   * Empty and singleton sets
   *)

  Definition empty : t := empty.
  Definition Empty (S : t) := forall x, ~ In x S.

  Lemma emptyIsEmpty:
    Empty empty.
  Proof.
    intros.
    apply empty_spec.
  Qed.

  Definition singleton (x : T.t) : t := singleton x.

  (**
   ** Properties
   *)

  Lemma emptySameAsEmpty:
    forall S, Empty S <-> S [=] empty.
  Proof.
    intros. split ; intro.
    {
      split ; intros.
      { contradict H0. apply H. }
      { contradict H0. apply emptyIsEmpty. }
    }
    { intro. intro. apply (emptyIsEmpty x). apply H. auto. }
  Qed.

  Lemma singletonNotEmpty:
    forall x, ~ singleton x [=] empty.
  Proof.
    intros.
    intro.
    pose proof H x.
    destruct H0.
    assert (In x (singleton x)).
    { apply singleton_spec ; auto. }
    pose proof H0 H2.
    contradict H3.
    apply emptyIsEmpty.
  Qed.

  Hint Resolve singletonNotEmpty : flam.

  Lemma singletonInv:
    forall x y, In x (singleton y) -> x = y.
  Proof.
    intros.
    unfold In, singleton in H.
    eapply singleton_spec.
    auto.
  Qed.

  Hint Resolve singletonInv : flam.
  
  Lemma singletonIntro:
    forall x y, x = y -> In x (singleton y).
  Proof.
    intros.
    unfold In, singleton.
    eapply singleton_spec.
    auto.
  Qed.

  Hint Resolve singletonInv : flam.

  Lemma lessThanEmpty:
    forall (A : t), included A empty -> A [=] empty.
  Proof.
    intros.
    apply empty_is_empty_1.
    intro. intro.
    contradict H.
    intro.
    unfold included, Subset in H.
    pose proof H a H0.
    contradict H1.
    apply emptyIsEmpty.
  Qed.

  Lemma emptyMinimal:
    forall A, included empty A.
  Proof.
    intros.
    apply subset_empty.
  Qed.

  Hint Resolve emptyMinimal : flam.

  Instance emptyProper:
    Proper (sameSet ==> iff) Empty.
  Proof.
    intro. intros.
    split ; intros.
    {
      intro. intro.
      pose proof H0 x0.
      contradict H2.
      apply H.
      auto.
    }
    {
      intro. intro.
      pose proof H0 x0.
      contradict H2.
      apply H.
      auto.
    }
  Qed.

  (**
   * Union
   *)

  Definition union (A B : t) : t := union A B.

  (**
   ** Properties
   *)

  Lemma unionIntro:
    forall (A B : t) x,
      In x A \/ In x B -> In x (union A B).
  Proof.
    eapply union_spec.
  Qed.

  Hint Resolve unionIntro : flam.

  Lemma unionInv:
    forall (A B : t) x,
      In x (union A B)
        -> In x A \/ In x B.
  Proof.
    eapply union_spec.
  Qed.

  Hint Resolve unionInv : flam.

  Lemma unionInv3 : forall d D1 D2 D3,
    In d (union D1 (union D2 D3))
      -> In d D1 \/ In d D2 \/ In d D3.
  Proof.
    intros.
    pose proof unionInv _ _ _ H.
    destruct H0.
      auto.
      pose proof unionInv _ _ _ H0.
      auto.
  Qed.

  Lemma emptyZero:
    forall (A : t), union empty A [=] A.
  Proof.
    intros.
    apply empty_union_1.
    apply empty_spec.
  Qed.

  Hint Resolve emptyZero : flam.

  Lemma emptyZero':
    forall (A : t), A [=] union empty A.
  Proof.
    intros. symmetry. apply emptyZero.
  Qed.

  Hint Resolve emptyZero' : flam.

  Lemma EmptyZero:
    forall (A B : t), Empty A -> union A B [=] B.
  Proof.
    intros.
    apply empty_union_1.
    auto.
  Qed.

  Hint Resolve EmptyZero : flam.

  Lemma unionComm:
    forall (A B : t),
      union A B [=] union B A.
  Proof.
    apply union_sym.
  Qed.

  Lemma unionAssoc:
    forall (A B C : t),
      union (union A B) C [=] union A (union B C).
  Proof.
    apply union_assoc.
  Qed.

  Hint Resolve unionAssoc : flam.

  Lemma unionIdemp:
    forall (A : t), union A A [=] A.
  Proof.
    intros.
    apply union_subset_equal.
    apply inclusionRefl.
  Qed.

  Lemma unionIncreasesL:
    forall (A B : t), included A (union A B).
  Proof.
    apply union_subset_1.
  Qed.

  Hint Resolve unionIncreasesL : flam.

  Lemma unionIncreasesR:
    forall (A B : t), included B (union A B).
  Proof.
    apply union_subset_2.
  Qed.

  Hint Resolve unionIncreasesR : flam.

  Lemma unionAbsorbs:
    forall (A B : t), included B A -> union A B [=] A.
  Proof.
    intros.
    rewrite unionComm.
    apply union_subset_equal.
    auto.
  Qed.

  Lemma unionMinimal:
    forall (A B C : t),
      included A C -> included B C -> included (union A B) C.
  Proof.
    apply union_subset_3.
  Qed.

  Hint Resolve unionMinimal : flam.

  Instance unionProper:
    Proper (sameSet ==> sameSet ==> sameSet) union.
  Proof.
    intro. intros. intro. intros.
    split ; intro.
    {
      destruct (unionInv _ _ _ H1) ; apply unionIntro.
      { left. apply H. auto. }
      { right. apply H0. auto. }
    }
    {
      destruct (unionInv _ _ _ H1) ; apply unionIntro.
      { left. apply H. auto. }
      { right. apply H0. auto. }
    }
  Qed.

  (**
   * Minus
   *)

  Definition minus (A B : t) : t := diff A B.

  (**
   ** Properties
   *)

  Lemma minusIntro:
    forall (A B : t) x,
      In x A -> ~ In x B -> In x (minus A B).
  Proof.
    intros.
    apply diff_spec.
    auto.
  Qed.

  Lemma minusInv:
    forall (A B : t) x,
      In x (minus A B)
        -> In x A /\ ~ In x B.
  Proof.
    apply diff_spec.
  Qed.

  Hint Resolve minusInv : flam.

  Instance minusProper:
    Proper (sameSet ==> sameSet ==> sameSet) minus.
  Proof.
    intro. intros. intro. intros.
    split ; intro.
    {
      destruct (minusInv _ _ _ H1).
      apply minusIntro.
      { apply H. auto. }
      { intro. contradict H3. apply H0. auto. }
    }
    {
      destruct (minusInv _ _ _ H1).
      apply minusIntro.
      { apply H. auto. }
      { intro. contradict H3. apply H0. auto. }
    }
  Qed.

  (**
   * Intersection
   *)

  Definition intersect (A B : t) : t := inter A B.

  (**
   ** Properties
   *)

  Lemma intersectInv:
    forall (A B : t) x,
      In x (intersect A B)
        -> In x A /\ In x B.
  Proof.
    apply inter_spec.
  Qed.

  Hint Resolve intersectInv : flam.

  Lemma intersectIntro:
    forall (A B : t) x,
      In x A -> In x B -> In x (intersect A B).
  Proof.
    intros. apply inter_spec. auto.
  Qed.

  Instance intersectProper:
    Proper (sameSet ==> sameSet ==> sameSet) intersect.
  Proof.
    intro. intros. intro. intros.
    split ; intros.
    {
      destruct (intersectInv _ _ _ H1).
      apply intersectIntro.
      { apply H. auto. }
      { apply H0. auto. }
    }
    {
      destruct (intersectInv _ _ _ H1).
      apply intersectIntro.
      { apply H. auto. }
      { apply H0. auto. }
    }
  Qed.

  (**
   * Disjoint sets
   *)

  Definition disjoint (A B : t) := Empty (intersect A B).

  (**
   ** Properties
   *)

  Lemma disjointIntro:
    forall (A B : t),
      (forall x, ~In x (intersect A B))
        -> disjoint A B.
  Proof.
    intros. intro. apply H.
  Qed.

  Hint Resolve disjointIntro : flam.

  Instance disjointProper:
    Proper (sameSet ==> sameSet ==> iff) disjoint.
  Proof.
    intro. intros. intro. intros.
    split ; intros.
    {
      unfold disjoint.
      rewrite <- H, <- H0.
      auto.
    }
    {
      unfold disjoint.
      rewrite H, H0.
      auto.
    }
  Qed.

  (**
   * Fold
   *)

  Section FOLD.
    Variable B : Type.
    Variable eqB : B -> B -> Prop.
    Variable f : T.t -> B -> B.

    Definition idempotent :=
      forall x init, eqB (f x (f x init)) (f x init).

    Definition fold (D : t) (init : B) : B :=
      fold f D init.

    (**
     ** Properties
     *)

    (* In particular, eqB is an equivalence relation *)
    Variable st : Equivalence eqB.

    Notation "b1 =B= b2" := (eqB b1 b2) (at level 70, no associativity).

    (* f is proper *)
    Variable Comp : Proper (Logic.eq ==> eqB ==> eqB) f.

    (* f is commutative: f x (f y init) =B= (f x init) *)
    Variable Trans : transpose eqB f.

    (* f is idempotent *)
    Variable Idemp : idempotent.

    Instance foldProper:
      Proper (sameSet ==> eqB ==> eqB) fold.
    Proof.
      intros.
      intro. intros. intro. intros.
      assert (fold x x0 =B= fold x y0).
      { eapply fold_init with (A := B) (eqA := eqB) ; auto. }

      assert (fold x y0 =B= fold y y0).
      { eapply fold_equal with (A := B) (eqA := eqB) ; auto. }

      destruct st. eauto.
    Qed.

    (* Induction principle for fold *)
    Theorem fold_rec:
      forall (P : t -> B -> Type) (init : B) (S : t),
         (forall S, Empty S -> P S init)
      -> (forall (x : T.t) (accum : B) (S1 S2 : t),
               In x S
            -> ~ In x S1
            -> add x S1 S2
            -> P S1 accum
            -> P S2 (f x accum))
      -> P S (fold S init).
    Proof.
      intros.
      unfold fold.
      apply fold_rec ; auto.
    Qed.

    Lemma foldEmpty:
      forall (init : B) (S : t),
        Empty S -> fold S init = init.
    Proof.
      intros.
      apply fold_1b.
      auto.
    Qed.

    Lemma foldSingleton:
      forall (init : B) (x : T.t),
        fold (singleton x) init =B= f x init.
    Proof.
      intros.
      rewrite singleton_equal_add.
      unfold fold.
      rewrite (fold_add st Comp Trans init).
      { rewrite foldEmpty. apply st. apply empty_spec. }
      { apply empty_spec. }
    Qed.

    Lemma foldAdd:
      forall (S1 S2 : t) (x : T.t) (init : B),
        ~ In x S1 ->
        add x S1 S2 ->
        fold S2 init =B= f x (fold S1 init).
    Proof.
      intros.
      unfold fold.
      apply fold_2 ; auto.
    Qed.

    Lemma foldInclude:
      forall (S1 S2 : t) (init : B),
        included S1 S2
          -> fold S2 init =B= fold (minus S2 S1) (fold S1 init).
    Proof.
      intros.
      unfold fold.
      assert (inter S2 S1 [=] S1) as Hrewrite.
      {
        rewrite inter_sym.
        apply inter_subset_equal.
        auto.
      }

      rewrite <- Hrewrite at 2.
      unfold minus.
      rewrite fold_diff_inter ; eauto.
      apply st.
    Qed.

    Lemma foldComm:
      forall (S1 S2 : t) (init : B),
        fold S1 (fold S2 init) =B= fold S2 (fold S1 init).
    Proof.
      intro.
      induction S1 using set_induction ; intros.
      {
        do 2 (rewrite (foldEmpty _ S1) ; auto).
        apply st.
      }
      {
        do 2 (rewrite (foldAdd S1_1 S1_2 x _) ; auto).
        rewrite IHS1_1.
        rewrite (fold_commutes st Comp Trans _ S2 x).
        apply st.
      }
    Qed.

    Lemma foldUnion:
      forall (S1 S2 : t) (init : B),
        fold (union S1 S2) init =B= fold S1 (fold S2 init).
    Proof.
      assert (forall S x init,
          In x S -> fold S init =B= f x (fold S init)) as HelpIn.
      {
        induction S using set_induction ; intros.
        { contradict H0. apply H. }
        {
          rewrite foldAdd ; eauto.
          rewrite Trans.
          destruct (H0 x0) as [Hadd _].
          destruct (Hadd H1).
          {
            subst x0.
            unfold idempotent in Idemp.
            rewrite Idemp.
            apply st.
          }
          { rewrite <- (IHS1 x0) ; auto. apply st. }
        }
      }

      assert (forall S0 x S init,
          add x S0 S -> fold S init =B= f x (fold S0 init)) as Help.
      {
        intros.
        destruct (In_dec x S0).
        {
          assert (S [=] S0).
          {
            intro. split ; intros.
            {
              destruct (H x0) as [Hadd _].
              destruct (Hadd H0) ; subst ; auto.
            }
            { apply H. auto. }
          }

          rewrite H0.
          rewrite <- HelpIn ; eauto.
          apply st.
        }
        { rewrite foldAdd ; eauto. apply st. }
      }

      induction S1 using set_induction ; intros.
      {
        destruct (emptySameAsEmpty S1) as [Hrewrite _].
        rewrite Hrewrite ; auto.
        rewrite emptyZero.
        rewrite (foldEmpty _ empty).
        { apply st. }
        { apply emptyIsEmpty. }
      }
      {
        rewrite (foldAdd S1_1 S1_2) ; eauto.
        rewrite (Help (union S1_1 S2) x _).
        { rewrite IHS1_1. apply st. }
        {
          intro. split ; intros.
          {
            destruct (unionInv _ _ _ H1).
            {
              destruct (H0 x') as [Hadd _].
              destruct (Hadd H2) ; auto.
              right. apply unionIntro. auto.
            }
            { right. apply unionIntro. auto. }
          }
          {
            destruct H1.
            { subst. apply unionIntro. left. apply H0. auto. }
            {
              destruct (unionInv _ _ _ H1).
              { apply unionIntro. left. apply H0. auto. }
              { apply unionIntro. auto. }
            }
          }
        }
      }
    Qed.
  End FOLD.

  (**
   * Filter
   *)

  Section FILTER.
    Variable f : T.t -> bool.

    Definition filter (S : t) : t :=
      filter f S.

    (**
     ** Properties
     *)

    Lemma filterCorrect:
      forall x S, In x (filter S) <-> In x S /\ f x = true.
    Proof.
      intros.
      unfold filter.
      apply filter_spec.

      intro. intros. subst. auto.
    Qed.

  End FILTER.

  (**
   * Choose
   *)

  (*
   * Returns an element of the given set, or None if the set is empty.
   * The element returned is unspecified, and equivalent sets may reurn
   * different elements.
   *)
  Definition choose (S : t) : option T.t := choose S.

  (**
   ** Properties
   *)

  Lemma chooseCorrect1:
    forall S x, choose S = Some x -> In x S.
  Proof.
    intros. unfold choose in H. apply choose_spec1. auto.
  Qed.

  Lemma chooseCorrect2:
    forall S, choose S = None -> Empty S.
  Proof.
    intros. unfold choose in H. apply choose_spec2. auto.
  Qed.

  (* begin hide *)
  (**************************************************************************)
  (* end hide *)

  (**
   * Miscellaneous properties
   *)

  Lemma set_ind:
    forall P : t -> Type,
         (forall S : t, Empty S -> P S)
      -> (forall S S' : t,
           P S -> forall x : T.t, ~In x S -> add x S S' -> P S')
      -> forall S : t, P S.
  Proof.
    apply set_induction.
  Qed.

  Lemma disjointEmpty:
    forall A, disjoint A empty /\ disjoint empty A.
  Proof.
    intros.
    split.
    {
      apply disjointIntro.
      intros.
      intro.
      assert (In x empty).
      { eapply inter_spec. eauto. }

      contradict H0.
      apply empty_spec.
    }
    {
      apply disjointIntro.
      intros.
      intro.
      assert (In x empty).
      { eapply inter_spec. rewrite inter_sym. eauto. }

      contradict H0.
      eapply empty_spec.
    }
  Qed.

  Hint Resolve disjointEmpty : flam.

  Lemma disjointMinus:
    forall (A B : t), disjoint A B -> minus A B [=] A.
  Proof.
    intros.
    split ; intros.
    {
      destruct (minusInv _ _ _ H0).
      auto.
    }
    {
      apply minusIntro ; auto.
      intro.
      pose proof H x.
      contradict H2.
      apply intersectIntro ; auto.
    }
  Qed.

  Hint Resolve disjointMinus : flam.

  Lemma minusEmpty:
    forall A, minus A empty [=] A.
  Proof.
    intros.
    apply disjointMinus.
    apply disjointEmpty.
  Qed.

  Lemma minusDistUnion:
    forall (A B C : t),
      minus (union A B) C
        [=] union (minus A C) (minus B C).
  Proof.
    intros.
    split ; intros.
    {
      apply unionIntro.
      destruct (minusInv _ _ _ H).
      destruct (unionInv _ _ _ H0);
        [left|right];
        apply minusIntro;
        auto.
    }
    {
      destruct (unionInv _ _ _ H);
        destruct (minusInv _ _ _ H0);
        apply minusIntro;
        auto;
        apply unionIntro;
        auto.
    }
  Qed.

  Hint Resolve minusDistUnion : flam.

  Lemma minusMinus:
    forall (A B C : t),
      minus (minus A B) C [=] minus A (union B C).
  Proof.
    intros.
    split ; intros.
    {
      destruct (minusInv _ _ _ H).
      destruct (minusInv _ _ _ H0).
      apply minusIntro ; auto.
      intro.
      destruct (unionInv _ _ _ H4) ; tauto.
    }
    {
      destruct (minusInv _ _ _ H).
      apply minusIntro.
      {
        apply minusIntro ; auto.
        intro.
        contradict H1.
        apply unionIntro ; auto.
      }
      {
        intro.
        contradict H1.
        apply unionIntro ; auto.
      }
    }
  Qed.

  Hint Resolve minusMinus : flam.

  Lemma minusUnion:
    forall (A B : t),
      union (minus A B) B [=] union A B.
  Proof.
    intros.
    split ; intros ; destruct (unionInv _ _ _ H).
    {
      destruct (minusInv _ _ _ H0).
      apply unionIntro. auto.
    }
    { apply unionIntro. auto. }
    {
      destruct (In_dec x B).
      { apply unionIntro. auto. }
      {
        apply unionIntro.
        left.
        apply minusIntro ; auto.
      }
    }
    { apply unionIntro ; auto. }
  Qed.

  Hint Resolve minusUnion : flam.

  Lemma emptyUnion:
    forall (A B : t),
      union A B [=] empty -> A [=] empty.
  Proof.
    intros.
    split ; intros.
    {
      destruct (H x).
      apply H1.
      apply unionIntro ; auto.
    }
    {
      contradict H0.
      apply emptyIsEmpty.
    }
  Qed.

  Lemma emptyUnion2:
    forall D1 D2, empty [=] union D1 D2 -> D1 [=] empty /\ D2 [=] empty.
  Proof.
    intros.
    split ; [|rewrite unionComm in H] ; eapply emptyUnion ; symmetry ; eauto.
  Qed.

  Lemma emptyUnion3:
    forall D1 D2 D3,
      empty [=] union D1 (union D2 D3)
        -> D1 [=] empty /\ D2 [=] empty /\ D3 [=] empty.
  Proof.
    intros.
    pose proof emptyUnion2 D1 (union D2 D3) H.
    destruct H0.
    split ; [auto | eapply emptyUnion2 ; symmetry ; auto ].
  Qed.

  Lemma emptyUnion4:
    forall D1 D2 D3 D4,
      empty [=] union D1 (union D2 (union D3 D4))
        -> D1 [=] empty /\ D2 [=] empty /\ D3 [=] empty /\ D4 [=] empty.
  Proof.
    intros.
    pose proof emptyUnion2 D1 (union D2 (union D3 D4)) H.
    destruct H0.
    split ; [auto | eapply emptyUnion3 ; symmetry ; auto].
  Qed.

  Lemma addIsUnion:
    forall x A B,
      add x A B <-> B [=] union (singleton x) A.
  Proof.
    apply Add_Equal.
  Qed.

  Lemma unionAdd:
    forall x A B C,
      add x A B -> add x (union A C) (union B C).
  Proof.
    intros. apply union_Add. auto.
  Qed.

  Hint Resolve emptyUnion : flam.

  Lemma inclusion2in3:
    forall D1 D2 D3,
      included D2 (union D1 (union D2 D3)).
  Proof.
    intros.
    apply inclusionTrans with (y := union D2 D3).
      apply unionIncreasesL.
      apply unionIncreasesR.
  Qed.

  Hint Resolve inclusion2in3 : flam.

  Lemma inclusion3in3:
    forall D1 D2 D3,
      included D3 (union D1 (union D2 D3)).
  Proof.
    intros.
    rewrite (unionComm D2 D3).
    apply inclusion2in3.
  Qed.

  Hint Resolve inclusion3in3 : flam.

  Lemma inclusion2in4:
    forall D1 D2 D3 D4,
      included D2 (union D1 (union D2 (union D3 D4))).
  Proof.
    intros.
    apply inclusionTrans with (y := union D2 (union D3 D4)).
      apply unionIncreasesL.
      apply unionIncreasesR.
  Qed.

  Hint Resolve inclusion2in4 : flam.

  Lemma inclusion3in4:
    forall D1 D2 D3 D4,
      included D3 (union D1 (union D2 (union D3 D4))).
  Proof.
    intros.
    rewrite <- (unionAssoc D2 D3 D4).
    rewrite (unionComm D2 D3).
    rewrite (unionAssoc D3 D2 D4).
    apply inclusion2in4.
  Qed.

  Hint Resolve inclusion3in4 : flam.

  Lemma inclusion4in4:
    forall D1 D2 D3 D4,
      included D4 (union D1 (union D2 (union D3 D4))).
  Proof.
    intros.
    rewrite (unionComm D3 D4).
    apply inclusion3in4.
  Qed.

  Hint Resolve inclusion4in4 : flam.
End FiniteSet.
